try:
    import pandas as pd
except ImportError as e:
    raise ImportError("Please install with geff[pandas] to use this module.") from e

import warnings
from pathlib import Path

from zarr.storage import StoreLike

from geff.core_io._base_read import read_to_memory


def geff_to_dataframes(
    store: StoreLike,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Convert a GEFF store to a pandas DataFrame.

    Properties with more than 2 dimensions cannot be converted and will be skipped
    Properties with two dimensions where the second dimension is > 1 will be unpacked
    into separate columns with the name "{prop_name}_{dim_index}"

    Args:
        store (StoreLike): The store to convert.

    Returns:
        tuple[pd.DataFrame, pd.DataFrame]: The nodes and edges dataframes.
    """
    memory_geff = read_to_memory(store)
    dataframes = []

    # Construct dictionaries to convert to dataframes
    for data_type in ["node", "edge"]:
        df_dict = {}

        # Extract ids
        if data_type == "node":
            df_dict["id"] = memory_geff["node_ids"]
        else:
            df_dict["source"] = memory_geff["edge_ids"][:, 0]
            df_dict["target"] = memory_geff["edge_ids"][:, 1]

        # Conditional necessary to making typing happy :/
        if data_type == "node":
            props = memory_geff["node_props"]
        else:
            props = memory_geff["edge_props"]

        for name, prop in props.items():
            missing = prop["missing"]
            # Squeeze out any singleton dimensions
            values = prop["values"].squeeze()
            ndim = len(values.shape)
            if ndim == 2:
                # After squeezing out singleton dimensions, second dim must be > 1
                for i in range(values.shape[1]):
                    series = pd.Series(values[:, i])
                    if missing is not None and any(missing):
                        series.mask(missing, inplace=True)
                    df_dict[f"{name}_{i}"] = series

            elif ndim > 2:
                warnings.warn(
                    f"{data_type} {name} ({ndim}D) will not be exported to csv "
                    "with more than 2 dimensions",
                    stacklevel=2,
                )
                continue
            else:
                # Data is 1d
                series = pd.Series(values)
                if missing is not None and any(missing):
                    series.mask(missing, inplace=True)
                df_dict[name] = series

        dataframes.append(pd.DataFrame(df_dict))

    return tuple(dataframes)


def geff_to_csv(store: StoreLike, outpath: Path | str, overwrite: bool = False) -> None:
    """Convert a geff store to two csvs of nodes and edges

    Properties with more than 2 dimensions cannot be exported and will be skipped
    Properties with two dimensions where the second dimension is > 1 will be unpacked
    into separate columns with the name "{prop_name}_{dim_index}"

    Args:
        store (StoreLike): Path to store or StoreLike object
        outpath (Path | str): Path to save output csvs. Any file extension will be
            stripped and replaced with "-nodes.csv" and "-edges.csv"
        overwrite (bool): If true, existing csvs will be overwritten
    """
    # Convert to path and remove any existing suffix
    outpath = Path(outpath).with_suffix("")
    # Add node/edge.csv to path
    node_path = f"{outpath}-nodes.csv"
    edge_path = f"{outpath}-edges.csv"

    # Convert and write to disk
    node_df, edge_df = geff_to_dataframes(store)
    mode = "w" if overwrite else "x"
    node_df.to_csv(node_path, mode=mode)
    edge_df.to_csv(edge_path, mode=mode)



from __future__ import annotations

import os
import shutil
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Literal, TypeVar

import numpy as np
import zarr

from geff import _path

if TYPE_CHECKING:
    from collections.abc import Sequence

    from numpy.typing import ArrayLike
    from zarr.storage import StoreLike

    from geff._typing import PropDictNpArray

    T = TypeVar("T")

from urllib.parse import urlparse


def is_remote_url(path: str) -> bool:
    """Returns True if the path is a remote URL (http, https, ftp, sftp), otherwise False.

    Parameters
    ----------
    path : str
        path to a local or remote resource

    Returns
    -------
    bool
        True if the path is a remote URL, False otherwise
    """
    parsed = urlparse(path)
    return parsed.scheme in ("http", "https", "ftp", "sftp")


def remove_tilde(store: StoreLike) -> StoreLike:
    """
    Remove tilde from a store path/str, because zarr (3?) will not recognize
        the tilde and write the zarr in the wrong directory.

    Args:
        store (str | Path | zarr store): The store to remove the tilde from

    Returns:
        StoreLike: The store with the tilde removed
    """
    if isinstance(store, str | Path):
        store_str = str(store)
        if "~" in store_str:
            store = os.path.expanduser(store_str)
    return store


def open_storelike(store: StoreLike) -> zarr.Group:
    """Opens a StoreLike input as a zarr group

    Args:
        store (str | Path | zarr store): str/Path/store for a geff zarr

    Raises:
        FileNotFoundError: Path does not exist
        ValueError: store must be a zarr StoreLike

    Returns:
        zarr.Group: Opened zarr group
    """
    # Check if path exists for string/Path inputs
    if isinstance(store, str | Path):
        store_path = Path(store)
        if not is_remote_url(str(store_path)) and not store_path.exists():
            raise FileNotFoundError(f"Path does not exist: {store}")

    # Check for zarr spec v3 files being opened with zarr python v2 and warn if so
    if zarr.__version__.startswith("2"):
        spec_version = _detect_zarr_spec_version(store)
        if spec_version == 3:
            warnings.warn(
                "Attempting to open a zarr spec v3 file with zarr-python v2. "
                "This may cause compatibility issues. Consider upgrading to zarr-python v3 "
                "or recreating the file with zarr spec v2.",
                UserWarning,
                stacklevel=2,
            )

    # Open the zarr group from the store
    try:
        graph_group = zarr.open_group(store, mode="r")
    except Exception as e:
        raise ValueError(f"store must be a zarr StoreLike: {e}") from e

    return graph_group


# -----------------------------------------------------------------------------#
# helpers
# -----------------------------------------------------------------------------#


def expect_array(parent: zarr.Group, key: str, parent_name: str = "array") -> zarr.Array:
    """Return an array in the parent group with the given key, or raise ValueError."""
    arr = parent.get(key)
    if not isinstance(arr, zarr.Array):
        raise ValueError(f"{parent_name!r} group must contain an {key!r} array")
    return arr


def expect_group(parent: zarr.Group, key: str, parent_name: str = "graph") -> zarr.Group:
    """Return a group in the parent group with the given key, or raise ValueError."""
    grp = parent.get(key)
    if not isinstance(grp, zarr.Group):
        raise ValueError(f"{parent_name!r} group must contain a group named {key!r}")
    return grp


def _detect_zarr_spec_version(store: StoreLike) -> int | None:
    """Detect the zarr specification version of an existing zarr store.

    Args:
        store: The zarr store path or object

    Returns:
        int | None: The zarr spec version (2 or 3) if detectable, None if unknown
    """
    try:
        if isinstance(store, str | Path):
            store_path = Path(store)
            # Check for zarr v3 indicator: zarr.json instead of .zarray/.zgroup
            if (store_path / "zarr.json").exists():
                return 3
            # Check for zarr v2 indicators
            elif (store_path / ".zgroup").exists() or (store_path / ".zarray").exists():
                return 2
        else:
            # For store objects, try to detect based on metadata
            group = zarr.open_group(store, mode="r")
            # In zarr v3, metadata is stored differently
            if group.metadata.zarr_format == 3:  # pyright: ignore
                return 3
            elif group.metadata.zarr_format == 2:  # pyright: ignore
                return 2
    except Exception:
        # If we can't detect, return None
        pass

    return None


def setup_zarr_group(store: StoreLike, zarr_format: Literal[2, 3] = 2) -> zarr.Group:
    """Set up and return a zarr group for writing.

    Args:
        store: The zarr store path or object
        zarr_format: The zarr format version to use

    Returns:
        The opened zarr group
    """
    store = remove_tilde(store)

    # Check for trying to write zarr spec v3 with zarr python v2 and warn if so
    if zarr_format == 3 and zarr.__version__.startswith("2"):
        warnings.warn(
            "Requesting zarr spec v3 with zarr-python v2. "
            "zarr-python v2 does not support spec v3. "
            "Ignoring zarr_format=3 and writing zarr spec v2 instead. "
            "Consider upgrading to zarr-python v3 to write zarr spec v3 files.",
            UserWarning,
            stacklevel=2,
        )

    # open/create zarr container
    if zarr.__version__.startswith("3"):
        return zarr.open_group(store, mode="a", zarr_format=zarr_format)
    else:
        return zarr.open_group(store, mode="a")


def _get_common_type_dims(arr_seq: Sequence[ArrayLike | None]) -> tuple[np.dtype, int]:
    """Get a common dtype and number of dimensions that will work for all elements.

    Will use native numpy casting to unify all the dtypes, and will take the maximum
    number of dimensions of the elements. Nones are ignored.

    Args:
        arr_seq (Sequence[ArrayLike | None]): A sequence of array-like elements and Nones
            to infer the dtype and number of dimensions for.

    Raises:
        ValueError: If the elements do not have any shared dtype that can be cast to safely.

    Returns:
        tuple[np.dtype, int]: A dtype that all non-None elements can be cast to, and the
            maximum number of dimensions of any non-None element
    """
    ndim = None
    dtype = None

    for arr in arr_seq:
        if arr is None:
            continue
        element = np.asarray(arr)
        if ndim is None:
            ndim = element.ndim
            dtype = element.dtype
        else:
            ndim = max(element.ndim, ndim)
            if np.can_cast(dtype, element.dtype):
                dtype = np.promote_types(dtype, element.dtype)
            else:
                raise ValueError(
                    "All elements must have compatible dtypes. Cannot"
                    f"cast {dtype} and {element.dtype}."
                )

    if dtype is None or ndim is None:
        warnings.warn(
            "Variable length property sequence does not have any non-None elements - "
            "using ndim=1 and dtype=int64",
            stacklevel=2,
        )
        dtype = np.dtype("int64")
        ndim = 1
    return dtype, ndim


def construct_var_len_props(arr_seq: Sequence[ArrayLike | None]) -> PropDictNpArray:
    """Converts a sequence of array like and None objects into a geff._typing.PropDictNpArray

    Creates a missing array with the indices of the None objects. Converts each element of
    the sequence into a numpy array. Prepends dummy dimensions to each element to ensure
    all elements have the same number of dimensions. Casts all arrays into a common dtype
    (if there is one). Turns the sequence into a numpy array with dtype object.

    Args:
        arr_seq (Sequence[ArrayLike | None]): A sequence of properties, with one entry
            per node or edge. Missing values are indicated by None entries.

    Returns:
        PropDictNpArray: A standardized version of the input properties where all entries
            are numpy arrays contained in an object array.
    """
    values_arr = np.empty(shape=(len(arr_seq),), dtype=np.object_)
    missing_arr = np.zeros(shape=(len(arr_seq),), dtype=np.bool_)

    dtype, ndim = _get_common_type_dims(arr_seq)

    for i, arr in enumerate(arr_seq):
        if arr is None:
            missing_arr[i] = 1
            empty_shape = tuple(0 for _ in range(ndim))
            default_val = np.empty(shape=empty_shape, dtype=dtype)
            values_arr[i] = default_val
        else:
            element = np.asarray(arr, dtype=dtype)
            # prepend dummy axes if needed
            while element.ndim < ndim:
                element = np.expand_dims(element, axis=0)
            values_arr[i] = element
    return {"values": values_arr, "missing": missing_arr if missing_arr.any() else None}


def delete_geff(store: StoreLike, zarr_format: Literal[2, 3] = 2) -> None:
    """Delete a geff after writing

    Tries to handle multiple StoreLike inputs and avoids deleting non-geff contents
    in the store

    Args:
        store (StoreLike): StoreLike geff that should be deleted
        zarr_format (Literal[2, 3], optional): Zarr format used to write input store. Defaults to 2.
    """
    root = setup_zarr_group(store, zarr_format=zarr_format)

    # Delete node and edge groups
    del root[_path.NODES]
    del root[_path.EDGES]

    # If the root is empty, try to delete the root zarr
    if len(list(root.keys())) == 0:
        # Handle Path or str storelike
        if isinstance(store, Path) or isinstance(store, str):
            shutil.rmtree(store)
        else:
            # Try to get a valid path from the store
            try:
                path = store.path  # type: ignore
                if os.path.exists(path):
                    shutil.rmtree(path)
            except AttributeError:
                warnings.warn(
                    "Cannot delete root zarr directory, but geff contents have been deleted",
                    stacklevel=2,
                )
                del root.attrs["geff"]
    else:
        warnings.warn(
            "Found non-geff members in zarr. Exiting without deleting root zarr.", stacklevel=2
        )
        # Delete geff metadata from attrs
        del root.attrs["geff"]


def check_for_geff(store: StoreLike, zarr_format: Literal[2, 3] = 2) -> bool:
    """Check a StoreLike for an existing geff and return True if already present

    Args:
        store (StoreLike): StoreLike to check for a geff
        zarr_format (Literal[2, 3], optional): Defaults to 2.

    Returns:
        bool: True if a geff already exists
    """
    if isinstance(store, Path):
        exists = store.exists()
    elif isinstance(store, str):
        exists = os.path.exists(store)
    # If store is already open, check for geff key in metadata
    else:
        root = setup_zarr_group(store, zarr_format=zarr_format)
        exists = "geff" in root.attrs

    return exists



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

import numpy as np

from ._errors import MissingDependencyError

try:
    import rustworkx as rx
except ImportError as e:
    raise MissingDependencyError(
        "This module requires rustworkx to be installed. "
        "Please install it with `pip install 'geff[rx]'`."
    ) from e


from geff.core_io import write_dicts
from geff_spec.utils import create_or_update_metadata, update_metadata_axes

from ._backend_protocol import Backend
from ._graph_adapter import GraphAdapter

if TYPE_CHECKING:
    from collections.abc import Sequence

    from numpy.typing import NDArray
    from zarr.storage import StoreLike

    from geff._typing import PropDictNpArray
    from geff_spec import AxisType, GeffMetadata


# NOTE: see _api_wrapper.py read/write/construct for docs
class RxBackend(Backend):
    @property
    def GRAPH_TYPES(self) -> tuple[type[rx.PyGraph], type[rx.PyDiGraph]]:
        return rx.PyGraph, rx.PyDiGraph

    @staticmethod
    def construct(
        metadata: GeffMetadata,
        node_ids: NDArray[Any],
        edge_ids: NDArray[Any],
        node_props: dict[str, PropDictNpArray],
        edge_props: dict[str, PropDictNpArray],
    ) -> rx.PyGraph | rx.PyDiGraph:
        metadata = metadata

        graph = rx.PyDiGraph() if metadata.directed else rx.PyGraph()
        graph.attrs = metadata.model_dump()

        # Add nodes with populated properties
        node_ids = node_ids.tolist()
        props_per_node: list[dict[str, Any]] = [{} for _ in node_ids]

        # Populate node properties first
        indices = np.arange(len(node_ids))

        for name, prop_dict in node_props.items():
            values = prop_dict["values"]
            if prop_dict["missing"] is not None:
                current_indices = indices[~prop_dict["missing"]]
                values = values[current_indices]
            else:
                current_indices = indices

            values = values.tolist()
            current_indices = current_indices.tolist()

            for idx, val in zip(current_indices, values, strict=True):
                props_per_node[idx][name] = val

        # Add nodes with their properties
        rx_node_ids = graph.add_nodes_from(props_per_node)

        # Create mapping from geff node id to rustworkx node index
        to_rx_id_map = dict(zip(node_ids, rx_node_ids, strict=False))

        # Add edges if they exist
        if len(edge_ids) > 0:
            # converting to local rx ids
            edge_ids = np.vectorize(to_rx_id_map.__getitem__)(edge_ids)
            # Prepare edge data with properties
            edges_data: list[dict[str, Any]] = [{} for _ in edge_ids]
            indices = np.arange(len(edge_ids))

            for name, prop_dict in edge_props.items():
                values = prop_dict["values"]
                if prop_dict["missing"] is not None:
                    current_indices = indices[~prop_dict["missing"]]
                    values = values[current_indices]
                else:
                    current_indices = indices

                values = values.tolist()
                current_indices = current_indices.tolist()

                for idx, val in zip(current_indices, values, strict=True):
                    edges_data[idx][name] = val

            # Add edges with their properties
            graph.add_edges_from(
                [(e[0], e[1], d) for e, d in zip(edge_ids, edges_data, strict=True)]
            )

        graph.attrs["to_rx_id_map"] = to_rx_id_map

        return graph

    @staticmethod
    def write(
        graph: rx.PyGraph | rx.PyDiGraph,
        store: StoreLike,
        metadata: GeffMetadata | None = None,
        axis_names: list[str] | None = None,
        axis_units: list[str | None] | None = None,
        axis_types: list[AxisType | None] | None = None,
        axis_scales: list[float | None] | None = None,
        scaled_units: list[str | None] | None = None,
        axis_offset: list[float | None] | None = None,
        zarr_format: Literal[2, 3] = 2,
        structure_validation: bool = True,
        node_id_dict: dict[int, int] | None = None,
    ) -> None:
        directed = isinstance(graph, rx.PyDiGraph)
        metadata = create_or_update_metadata(metadata=metadata, is_directed=directed)
        if axis_names is not None:
            metadata = update_metadata_axes(
                metadata, axis_names, axis_units, axis_types, axis_scales, scaled_units, axis_offset
            )

        if graph.num_nodes() == 0:
            # Handle empty graph case - still need to write empty structure
            node_data: list[tuple[int, dict[str, Any]]] = []
            edge_data: list[tuple[tuple[int, int], dict[str, Any]]] = []
            node_props: list[str] = []
            edge_props: list[str] = []

        else:
            # Prepare node data
            if node_id_dict is None:
                node_data = [
                    (i, data) for i, data in zip(graph.node_indices(), graph.nodes(), strict=False)
                ]
            else:
                node_data = [
                    (node_id_dict[i], data)
                    for i, data in zip(graph.node_indices(), graph.nodes(), strict=False)
                ]

            # Prepare edge data
            if node_id_dict is None:
                edge_data = [((u, v), data) for u, v, data in graph.weighted_edge_list()]
            else:
                edge_data = [
                    ((node_id_dict[u], node_id_dict[v]), data)
                    for u, v, data in graph.weighted_edge_list()
                ]

            node_props = list({k for _, data in node_data for k in data})
            edge_props = list({k for _, data in edge_data for k in data})

        write_dicts(
            geff_store=store,
            node_data=node_data,
            edge_data=edge_data,
            node_prop_names=node_props,
            edge_prop_names=edge_props,
            metadata=metadata,
            zarr_format=zarr_format,
            structure_validation=structure_validation,
        )

    @staticmethod
    def graph_adapter(graph: rx.PyGraph | rx.PyDiGraph) -> RxGraphAdapter:
        return RxGraphAdapter(graph)


class RxGraphAdapter(GraphAdapter):
    def __init__(self, graph: rx.PyGraph | rx.PyDiGraph) -> None:
        self.graph = graph

    def get_node_ids(self) -> Sequence[int]:
        return list(self.graph.node_indices())

    def get_edge_ids(self) -> Sequence[tuple[int, int]]:
        return list(self.graph.edge_list())

    def has_node_prop(self, name: str, node: int, metadata: GeffMetadata) -> bool:
        return name in self.graph[node]

    def get_node_prop(
        self,
        name: str,
        node: int,
        metadata: GeffMetadata,
    ) -> Any:
        return self.graph[node][name]

    def has_edge_prop(self, name: str, edge: tuple[int, int], metadata: GeffMetadata) -> bool:
        return name in self.graph.get_edge_data(*edge)

    def get_edge_prop(
        self,
        name: str,
        edge: tuple[int, int],
        metadata: GeffMetadata,
    ) -> Any:
        return self.graph.get_edge_data(*edge)[name]



import re
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import numpy as np
import pytest
import zarr
import zarr.storage

from geff import _path
from geff.core_io import construct_var_len_props, write_arrays
from geff.core_io._base_read import read_to_memory
from geff.testing._utils import check_equiv_geff
from geff.testing.data import (
    create_simple_2d_geff,
    create_simple_3d_geff,
    create_simple_temporal_geff,
)
from geff.validate.structure import validate_structure
from geff_spec import GeffMetadata, PropMetadata

if TYPE_CHECKING:
    from geff._typing import InMemoryGeff, PropDictNpArray


from geff.core_io._base_write import (
    dict_props_to_arr,
    write_dicts,
    write_props_arrays,
)


def _tmp_metadata():
    """Return minimal valid GeffMetadata object for tests."""
    return GeffMetadata(
        geff_version="0.0.1", directed=True, node_props_metadata={}, edge_props_metadata={}
    )


@pytest.fixture
def dict_data():
    data = [
        (0, {"num": 1, "str": "category"}),
        (127, {"num": 5, "str_arr": ["test", "string"]}),
        (1, {"num": 6, "num_arr": [1, 2]}),
    ]
    return data


class TestWriteArrays:
    @pytest.mark.parametrize("zarr_format", [2, 3])
    def test_write_arrays_basic(self, tmp_path: Path, zarr_format: Literal[2, 3]) -> None:
        """Test basic functionality of write_arrays with minimal data."""
        # Create test data
        geff_path = tmp_path / "test.geff"
        node_ids = np.array([1, 2, 3], dtype=np.uint32)
        edge_ids = np.array([[1, 2], [2, 3]], dtype=np.uint32)
        metadata = _tmp_metadata()

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
                message="Requesting zarr spec v3 with zarr-python v2.*",
            )
            # Call write_arrays
            write_arrays(
                geff_store=geff_path,
                node_ids=node_ids,
                node_props={},
                edge_ids=edge_ids,
                edge_props={},
                metadata=metadata,
                zarr_format=zarr_format,
            )

        # Verify the zarr group was created
        assert geff_path.exists()

        # Verify node and edge IDs were written correctly
        root = zarr.open_group(str(geff_path))
        assert "nodes/ids" in root
        assert "edges/ids" in root

        # Check the data matches
        np.testing.assert_array_equal(root["nodes/ids"][:], node_ids)
        np.testing.assert_array_equal(root["edges/ids"][:], edge_ids)

        # Check the data types match
        assert root["nodes/ids"].dtype == node_ids.dtype
        assert root["edges/ids"].dtype == edge_ids.dtype

        # Verify metadata was written
        assert "geff" in root.attrs
        assert root.attrs["geff"]["geff_version"] == "0.0.1"
        assert root.attrs["geff"]["directed"] is True

        validate_structure(geff_path)

    def test_write_in_mem_geff(self):
        store, _attrs = create_simple_3d_geff()
        in_mem_geff = read_to_memory(store)

        # Test writing
        new_store = zarr.storage.MemoryStore()
        write_arrays(new_store, **in_mem_geff)

        validate_structure(new_store)

    def test_write_arrays_rejects_disallowed_id_dtype(self, tmp_path) -> None:
        """write_arrays must fail fast for node/edge ids with unsupported dtype."""
        geff_path = tmp_path / "invalid_ids.geff"

        # float16 is currently not allowed by GEFF Spec
        node_ids = np.array([1, 2, 3], dtype=np.float16)
        edge_ids = np.array([[1, 2], [2, 3]], dtype=np.float16)

        with pytest.raises(TypeError, match="Node ids and edge ids must have int dtype"):
            write_arrays(
                geff_store=geff_path,
                node_ids=node_ids,
                node_props=None,
                edge_ids=edge_ids,
                edge_props=None,
                metadata=_tmp_metadata(),
            )

    def test_write_arrays_upcasts_disallowed_property_dtype(self, tmp_path) -> None:
        """write_arrays must fail fast if any property array has an unsupported dtype."""
        geff_path = tmp_path / "invalid_prop.geff"

        # ids are fine (int32)
        node_ids = np.array([1, 2, 3], dtype=np.int32)
        edge_ids = np.array([[1, 2], [2, 3]], dtype=np.int32)

        # property with disallowed dtype (float16)
        bad_prop_values = np.array([0.1, 0.2, 0.3], dtype=np.float16)
        node_props: dict[str, PropDictNpArray] = {
            "score": {"values": bad_prop_values, "missing": None}
        }

        with pytest.warns(
            UserWarning, match="Dtype float16 is being upcast to float32 for Java compatibility"
        ):
            write_arrays(
                geff_store=geff_path,
                node_ids=node_ids,
                node_props=node_props,
                edge_ids=edge_ids,
                edge_props=None,
                metadata=_tmp_metadata(),
            )

    @pytest.mark.parametrize(
        "data_func", [create_simple_2d_geff, create_simple_3d_geff, create_simple_temporal_geff]
    )
    @pytest.mark.parametrize("zarr_format", [2, 3])
    def test_simple_geffs(self, data_func, zarr_format, tmp_path):
        memory_geff: InMemoryGeff
        store, memory_geff = data_func()
        path = tmp_path / "test.geff"
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
                message="Requesting zarr spec v3 with zarr-python v2.*",
            )
            write_arrays(path, **memory_geff, zarr_format=zarr_format)
            check_equiv_geff(store, path)

    def test_invalid_geff(self, tmp_path):
        _, memory_geff = create_simple_2d_geff()
        path = tmp_path / "test.geff"
        # Add array with wrong size
        memory_geff["node_props"]["bad_size"] = {"values": np.zeros((2, 10)), "missing": None}

        with pytest.raises(
            ValueError,
            match=re.escape(
                "Node property 'bad_size' values has length 2, which "
                "does not match id length 10\nCannot write invalid geff."
            ),
        ):
            write_arrays(path, **memory_geff)

    def test_existing_geff(self):
        store, memory_geff = create_simple_2d_geff()
        with pytest.raises(
            FileExistsError, match=r"Found an existing geff present in `geff_store`."
        ):
            write_arrays(store, **memory_geff)

        # Try with overwrite
        new_store, new_geff = create_simple_3d_geff()
        with pytest.raises(
            UserWarning,
            match="Cannot delete root zarr directory, but geff contents have been deleted",
        ):
            write_arrays(store, **new_geff, overwrite=True)
            check_equiv_geff(store, new_store)

    def test_write_props_metadata(self):
        _, memory_geff = create_simple_3d_geff()

        # Define props metadata
        props_meta = {
            "x": PropMetadata(
                identifier="x",
                dtype=str(memory_geff["node_props"]["x"]["values"].dtype),
                unit="um",
                description="xaxis",
                name="X axis",
            ),
            "y": PropMetadata(
                identifier="y",
                dtype=str(memory_geff["node_props"]["y"]["values"].dtype),
                unit="um",
                description="yaxis",
                name="Y axis",
            ),
            "z": PropMetadata(
                identifier="z",
                dtype=str(memory_geff["node_props"]["z"]["values"].dtype),
                unit="um",
                description="zaxis",
                name="Z axis",
            ),
            "t": PropMetadata(
                identifier="t",
                dtype=str(memory_geff["node_props"]["t"]["values"].dtype),
                unit="um",
                description="taxis",
                name="t axis",
            ),
        }
        memory_geff["metadata"].node_props_metadata = props_meta

        store = zarr.storage.MemoryStore()
        write_arrays(store, **memory_geff)

        new_meta = GeffMetadata.read(store)
        assert new_meta.node_props_metadata == props_meta


@pytest.mark.parametrize(
    ("data_type", "expected"),
    [
        ("num", ([1, 5, 6], None)),
        ("str", (["category", "", ""], [0, 1, 1])),
        ("num_arr", ([[1, 2], [1, 2], [1, 2]], [1, 1, 0])),
        ("str_arr", ([["test", "string"], ["test", "string"], ["test", "string"]], [1, 0, 1])),
    ],
)
def test_dict_prop_to_arr(dict_data, data_type, expected) -> None:
    props_dict = dict_props_to_arr(dict_data, [data_type])
    values = props_dict[data_type]["values"]
    missing = props_dict[data_type]["missing"]
    ex_values, ex_missing = expected
    ex_values = np.array(ex_values)
    ex_missing = np.array(ex_missing, dtype=bool) if ex_missing is not None else None

    np.testing.assert_array_equal(missing, ex_missing)
    np.testing.assert_array_equal(values, ex_values)


class Test_write_dicts:
    def test_node_ids_not_int(self):
        store = zarr.storage.MemoryStore()
        node_data = [(float(node), {}) for node in range(10)]
        meta = GeffMetadata(directed=False, node_props_metadata={}, edge_props_metadata={})
        with pytest.raises(UserWarning, match=r"Node ids with dtype .* are being cast to uint"):
            write_dicts(store, node_data, [], [], [], meta)

            z = zarr.open(store)
            assert np.issubdtype(z[_path.NODE_IDS].dtype, np.unsignedinteger)

    def test_node_int_to_uint(self):
        store = zarr.storage.MemoryStore()
        node_data = [(node, {}) for node in range(10)]
        meta = GeffMetadata(directed=False, node_props_metadata={}, edge_props_metadata={})
        write_dicts(store, node_data, [], [], [], metadata=meta)

        z = zarr.open(store)
        assert np.issubdtype(z[_path.NODE_IDS].dtype, np.unsignedinteger)

    def test_negative_ids(self):
        store = zarr.storage.MemoryStore()
        node_data = [(-node, {}) for node in range(10)]
        meta = GeffMetadata(directed=False, node_props_metadata={}, edge_props_metadata={})
        with pytest.raises(ValueError, match="Cannot write a geff with node ids that are negative"):
            write_dicts(store, node_data, [], [], [], metadata=meta)


@pytest.mark.parametrize("group", ["nodes", "edges"])
class TestWritePropsArrays:
    def _helper_for_testing_expected(
        self,
        store,
        props_metadata,
        group,
        prop_name,
        varlength,
        prop_dtype,
        expected_values,
        expected_missing,
        expected_data,
    ):
        prop_meta = None
        for prop_meta in props_metadata:
            if prop_meta.identifier == prop_name:
                break
        assert prop_meta is not None
        assert np.issubdtype(np.dtype(prop_meta.dtype), prop_dtype)
        assert prop_meta.varlength == varlength
        root = zarr.open(store)
        group_path = _path.NODES if group == "nodes" else _path.EDGES
        written_values = root[group_path][_path.PROPS][prop_name]["values"]
        np.testing.assert_array_equal(written_values[:], expected_values)

        if expected_missing is None:
            assert "missing" not in root[group_path][_path.PROPS][prop_name].keys()
        else:
            written_missing = root[group_path][_path.PROPS][prop_name]["missing"]
            np.testing.assert_array_equal(written_missing[:], expected_missing)
        if expected_data is None:
            assert "data" not in root[group_path][_path.PROPS][prop_name].keys()
        else:
            written_data = root[group_path][_path.PROPS][prop_name]["data"]
            np.testing.assert_array_equal(written_data[:], expected_data)

    def test_int_prop(self, group):
        store = zarr.storage.MemoryStore()
        prop_name = "ints"
        varlength = False
        prop_dtype = np.uint16
        prop = {
            "values": np.array([1, 2, 3, 4], dtype=np.uint16),
            "missing": np.array([0, 0, 0, 1], dtype=bool),
        }
        expected_values = prop["values"]
        expected_missing = prop["missing"]
        expected_data = None

        props_metadata = write_props_arrays(store, group, {prop_name: prop})
        self._helper_for_testing_expected(
            store,
            props_metadata,
            group,
            prop_name,
            varlength,
            prop_dtype,
            expected_values,
            expected_missing,
            expected_data,
        )

    def test_unsquish_prop(self, group):
        store = zarr.storage.MemoryStore()
        prop_name = "prop_to_unsqush"
        new_prop1 = "s1"
        new_prop2 = "s2"
        varlength = False
        prop_dtype = np.float32
        prop = {
            "values": np.array([[1, 2], [3, 4]], dtype=prop_dtype),
            "missing": np.array([0, 1], dtype=bool),
        }
        expected_values1 = np.array([1, 3], dtype=prop_dtype)
        expected_values2 = np.array([2, 4], dtype=prop_dtype)
        expected_missing = prop["missing"]
        expected_data = None
        props_unsquish = {prop_name: [new_prop1, new_prop2]}
        props_metadata = write_props_arrays(
            store, group, {prop_name: prop}, props_unsquish=props_unsquish
        )

        prop_name = new_prop1
        expected_values = expected_values1
        for prop_name, expected_values in zip(
            [new_prop1, new_prop2], [expected_values1, expected_values2], strict=True
        ):
            self._helper_for_testing_expected(
                store,
                props_metadata,
                group,
                prop_name,
                varlength,
                prop_dtype,
                expected_values,
                expected_missing,
                expected_data,
            )

    def test_str_prop(self, group):
        store = zarr.storage.MemoryStore()
        prop_name = "str"
        varlength = False
        prop_dtype = "str"
        prop = {
            "values": np.array(["", "test", "str"], dtype=prop_dtype),
            "missing": np.array([0, 1, 0], dtype=bool),
        }
        expected_values = prop["values"]
        expected_missing = prop["missing"]
        expected_data = None
        props_metadata = write_props_arrays(store, group, {prop_name: prop})
        self._helper_for_testing_expected(
            store,
            props_metadata,
            group,
            prop_name,
            varlength,
            prop_dtype,
            expected_values,
            expected_missing,
            expected_data,
        )

    def test_str_array_prop(self, group):
        store = zarr.storage.MemoryStore()
        prop_name = "str_arr"
        varlength = False
        prop_dtype = "str"
        prop = {
            "values": np.array(
                [
                    [
                        "",
                        "test",
                    ],
                    ["str", "array"],
                ],
                dtype=np.str_,
            ),
            "missing": np.array([0, 1], dtype=bool),
        }
        expected_values = prop["values"]
        expected_missing = prop["missing"]
        expected_data = None
        props_metadata = write_props_arrays(store, group, {prop_name: prop})
        self._helper_for_testing_expected(
            store,
            props_metadata,
            group,
            prop_name,
            varlength,
            prop_dtype,
            expected_values,
            expected_missing,
            expected_data,
        )

    def test_varlength_1d(self, group):
        store = zarr.storage.MemoryStore()
        prop_name = "varlength"
        varlength = True
        prop_dtype = "int"
        prop = construct_var_len_props(
            [
                [1, 2, 3],
                [4, 5],
                None,
            ]
        )
        expected_values = np.array(
            [
                [
                    0,
                    3,
                ],
                [3, 2],
                [5, 0],
            ]
        )
        expected_missing = prop["missing"]
        expected_data = np.array([1, 2, 3, 4, 5], dtype="int")
        props_metadata = write_props_arrays(store, group, {prop_name: prop})

        self._helper_for_testing_expected(
            store,
            props_metadata,
            group,
            prop_name,
            varlength,
            prop_dtype,
            expected_values,
            expected_missing,
            expected_data,
        )

    def test_varlength_3d(self, group):
        store = zarr.storage.MemoryStore()
        prop_name = "varlength"
        varlength = True
        prop_dtype = "int"
        prop = construct_var_len_props([[[[0], [1], [2]]], [[4, 5], [6, 7]]])

        expected_values = np.array(
            [
                [0, 1, 3, 1],
                [3, 1, 2, 2],
            ]
        )
        expected_missing = prop["missing"]
        expected_data = np.array([0, 1, 2, 4, 5, 6, 7], dtype="int")
        props_metadata = write_props_arrays(store, group, {prop_name: prop})

        self._helper_for_testing_expected(
            store,
            props_metadata,
            group,
            prop_name,
            varlength,
            prop_dtype,
            expected_values,
            expected_missing,
            expected_data,
        )



import warnings

import networkx as nx
import networkx.algorithms.isomorphism as iso
from zarr.storage import StoreLike

from geff import _path
from geff.core_io._utils import expect_array, expect_group, open_storelike
from geff.validate.structure import validate_structure


def nx_is_equal(g1: nx.Graph, g2: nx.Graph) -> bool:
    """Utility function to check that two Network graphs are perfectly identical.

    It checks that the graphs are isomorphic, and that their graph,
    nodes and edges attributes are all identical.

    Args:
        g1 (nx.Graph): The first graph to compare.
        g2 (nx.Graph): The second graph to compare.

    Returns:
        bool: True if the graphs are identical, False otherwise.
    """
    edges_attr = list({k for (n1, n2, d) in g2.edges.data() for k in d})
    edges_default = len(edges_attr) * [0]
    em = iso.categorical_edge_match(edges_attr, edges_default)
    nodes_attr = list({k for (n, d) in g2.nodes.data() for k in d})
    nodes_default = len(nodes_attr) * [0]
    nm = iso.categorical_node_match(nodes_attr, nodes_default)

    same_nodes = same_edges = False
    if not g1.nodes.data() and not g2.nodes.data():
        same_nodes = True
    elif len(g1.nodes.data()) != len(g2.nodes.data()):
        same_nodes = False
    else:
        for data1, data2 in zip(sorted(g1.nodes.data()), sorted(g2.nodes.data()), strict=False):
            n1, attr1 = data1
            n2, attr2 = data2
            if sorted(attr1) == sorted(attr2) and n1 == n2:
                same_nodes = True
            else:
                same_nodes = False

    if not g1.edges.data() and not g2.edges.data():
        same_edges = True
    elif len(g1.edges.data()) != len(g2.edges.data()):
        same_edges = False
    else:
        for data1, data2 in zip(sorted(g1.edges.data()), sorted(g2.edges.data()), strict=False):
            n11, n12, attr1 = data1
            n21, n22, attr2 = data2
            if sorted(attr1) == sorted(attr2) and sorted((n11, n12)) == sorted((n21, n22)):
                same_edges = True
            else:
                same_edges = False

    if (
        nx.is_isomorphic(g1, g2, edge_match=em, node_match=nm)
        and g1.graph == g2.graph
        and same_nodes
        and same_edges
    ):
        return True
    else:
        return False


def check_equiv_geff(store_a: StoreLike, store_b: StoreLike) -> None:
    """This function compares two geffs, typically a starting fixture geff with
    the output of an implementation.

    This tests focuses on maintaining shape and dtype consistency. It does not
    assert element wise equality. store_a is assumed to be the "correct" geff.

    Missing arrays are not required to be present in both a and b because we allow
    missing arrays where all values are present. We raise a warning if we see this.

    Args:
        store_a (str | Path | zarr store): str/Path/store for a geff zarr
        store_b (str | Path | zarr store): str/Path/store for a second geff zarr
    """

    # Run validation first so that we don't hit issues with basic structure
    validate_structure(store_a)
    validate_structure(store_b)

    for graph_group in [_path.NODES, _path.EDGES]:
        ga = expect_group(open_storelike(store_a), graph_group)
        gb = expect_group(open_storelike(store_b), graph_group)

        # Check ids
        ids_a = expect_array(ga, _path.IDS)
        ids_b = expect_array(gb, _path.IDS)
        if (a_shape := ids_a.shape) != (b_shape := ids_b.shape):
            raise ValueError(f"{graph_group} ids shape: a {a_shape} does not match b {b_shape}")

        # Check that properties in each geff are the same
        a_props = set(expect_group(ga, _path.PROPS))
        b_props = set(expect_group(gb, _path.PROPS))
        if a_props != b_props:
            raise ValueError(
                f"{graph_group} properties: a ({a_props}) does not match b ({b_props})"
            )

        # Check shape and dtype of each prop
        for prop in a_props:
            a_missing = "missing" in expect_group(ga, f"{_path.PROPS}/{prop}")
            b_missing = "missing" in expect_group(gb, f"{_path.PROPS}/{prop}")
            if a_missing != b_missing:
                warnings.warn(
                    f"one {graph_group}/{_path.PROPS}/{prop} contains missing "
                    "but the other does not. This may be correct but should be verified.",
                    stacklevel=2,
                )

                # Note: don't need to check missing shape because validation forces it to be
                # the same shape as values

                # Note: don't need to check missing dtype because validation forces it to be bool

            a_values = expect_array(ga, f"{_path.PROPS}/{prop}/{_path.VALUES}")
            b_values = expect_array(gb, f"{_path.PROPS}/{prop}/{_path.VALUES}")
            if (a_shape := a_values.shape) != (b_shape := b_values.shape):
                raise ValueError(
                    f"{graph_group}/{_path.PROPS}/{prop}/{_path.VALUES} shape: "
                    f"a {a_shape} does not match b {b_shape}"
                )
            if (a_dtype := a_values.dtype) != (b_dtype := b_values.dtype):
                raise ValueError(
                    f"{graph_group}/{_path.PROPS}/{prop}/{_path.VALUES} dtype: "
                    f"a {a_dtype} does not match b {b_dtype}"
                )



import numpy as np

from geff.validate.graph import (
    validate_no_repeated_edges,
    validate_no_self_edges,
    validate_nodes_for_edges,
    validate_unique_node_ids,
)


def test_no_self_edges() -> None:
    """
    Test that no node has an edge to itself.
    """
    edge_ids = np.array([[0, 1], [1, 2], [2, 3]])
    is_valid, problematic_nodes = validate_no_self_edges(edge_ids)
    assert is_valid, "There should be no self-edges in the GEFF group."
    assert len(problematic_nodes) == 0, "There should be no problematic nodes with self-edges."


def test_detects_self_edges() -> None:
    """
    Test that validator detects nodes with self-edges.
    """
    edge_ids = np.array([[0, 1], [1, 2], [2, 3], [0, 0]])  # Node 0 has a self-edge
    is_valid, problematic_nodes = validate_no_self_edges(edge_ids)
    assert not is_valid, "Validator should detect self-edges."
    assert len(problematic_nodes) > 0, "There should be problematic nodes with self-edges."
    assert np.array_equal(problematic_nodes, np.array([0])), (
        "Node 0 should be the problematic node with a self-edge."
    )


def test_all_edges_valid() -> None:
    """
    Test that all edges reference existing node IDs.
    """
    node_ids = np.array([0, 1, 2, 3])
    edge_ids = np.array([[0, 1], [1, 2], [2, 3]])
    is_valid, invalid_edges = validate_nodes_for_edges(node_ids, edge_ids)
    assert is_valid, "All edges should reference valid node IDs."
    assert len(invalid_edges) == 0, "There should be no invalid edges."


def test_detects_invalid_edges() -> None:
    """
    Test that invalid edges (edges with missing node IDs) are detected.
    """
    node_ids = np.array([0, 1, 2])
    edge_ids = np.array([[0, 1], [1, 2], [2, 3]])
    is_valid, invalid_edges = validate_nodes_for_edges(node_ids, edge_ids)
    assert not is_valid, "Validator should detect edges referencing missing node IDs."
    assert (2, 3) in invalid_edges, "Edge (2, 3) should be flagged as invalid."
    assert len(invalid_edges) == 1, "There should be exactly one invalid edge."


def test_no_repeated_edges() -> None:
    """
    Test that validator passes when all edges are unique.
    """
    edge_ids = np.array([[0, 1], [1, 2], [2, 3]])
    is_valid, repeated_edges = validate_no_repeated_edges(edge_ids)
    assert is_valid, "There should be no repeated edges."
    assert len(repeated_edges) == 0, "No edges should be reported as repeated."


def test_detects_repeated_edges() -> None:
    """
    Test that validator detects repeated edges.
    """
    edge_ids = np.array([[0, 1], [1, 2], [2, 3], [0, 1]])  # Edge (0, 1) is repeated
    is_valid, repeated_edges = validate_no_repeated_edges(edge_ids)
    assert not is_valid, "Validator should detect repeated edges."
    assert [0, 1] in repeated_edges.tolist(), "Edge [0, 1] should be reported as repeated."
    assert len(repeated_edges) == 1, "There should be exactly one unique repeated edge."


def test_validate_unique_node_ids() -> None:
    node_ids = np.array([0, 0, 1, 2, 3])
    valid, nonunique = validate_unique_node_ids(node_ids)
    assert not valid
    assert nonunique == np.array([0])



import os
from pathlib import Path
from typing import Literal

from geff.core_io._utils import check_for_geff, delete_geff

try:
    import tifffile

    # TODO: would be nice to remove dask dependency just to create Zarr from Tiff
    # tifffile directly supports creating Zarr stores from sequences of Tiff files
    # The only convenience added is the ability to easily expand the data to (T, C, Z, Y, X)
    # in the case of images that cannot fit into memory.
    from dask.array.image import imread
    from skimage.measure import regionprops
except ImportError as e:
    raise ImportError("Please install with geff[ctc] to use this module.") from e

import warnings

import numpy as np
import zarr
from zarr.storage import StoreLike

import geff
from geff.core_io import write_arrays
from geff_spec import Axis, GeffMetadata, RelatedObject


def ctc_tiffs_to_zarr(
    ctc_path: Path,
    output_store: StoreLike,
    ctzyx: bool = False,
    overwrite: bool = False,
    zarr_format: Literal[2, 3] = 2,
) -> None:
    """
    Convert a CTC file to a Zarr file.

    Args:
        ctc_path (Path): The path to the CTC file.
        output_store (StoreLike): The path to the Zarr file.
        ctzyx (optional, bool): Expand data to make it (T, C, Z, Y, X)
            otherwise it's (T,) + Frame shape. Defaults to False.
        overwrite (optional, bool): Whether to overwrite the Zarr file if it already exists.
            Defaults to False.
        zarr_format (optional, Literal[2, 3]): The zarr specification to use when writing the zarr.
            Defaults to 2.
    """
    array = imread(str(ctc_path / "*.tif"))
    if ctzyx:
        n_missing_dims = 5 - array.ndim  # (T, C, Z, Y, X)
        expand_dims = (slice(None),) + (np.newaxis,) * n_missing_dims
        array = array[expand_dims]
    # Warning is triggered when the zarr_format argument is ignored by zarr 2
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=r".*ignoring keyword argument.*zarr_format.*", category=UserWarning
        )
        array.to_zarr(url=output_store, overwrite=overwrite, zarr_format=zarr_format)


def from_ctc_to_geff(
    ctc_path: Path,
    geff_path: Path,
    segmentation_store: StoreLike | None = None,
    tczyx: bool = False,
    overwrite: bool = False,
    zarr_format: Literal[2, 3] = 2,
) -> None:
    """
    Convert a CTC file to a GEFF file.

    Args:
        ctc_path: The path to the CTC file.
        geff_path: The path to the GEFF file.
        segmentation_store: The path or store to save the segmentation to.
                            If not provided, it won't be exported.
        tczyx: Expand data to make it (T, C, Z, Y, X) otherwise it's (T,) + Frame shape.
        overwrite: Whether to overwrite the GEFF file if it already exists.
        zarr_format (Literal[2, 3]): The zarr specification to use when writing the zarr.
            Defaults to 2.
    """
    ctc_path = Path(ctc_path)
    geff_path = Path(geff_path).with_suffix(".geff")

    if not ctc_path.exists():
        raise FileNotFoundError(f"CTC file {ctc_path} does not exist")

    for tracks_file in ["man_track.txt", "res_track.txt"]:
        tracks_file_path = ctc_path / tracks_file
        if tracks_file_path.exists():
            break
    else:
        raise FileNotFoundError(
            f"Tracks file {ctc_path}/man_track.txt or {ctc_path}/res_track.txt does not exist"
        )

    # Check for existing geff
    if check_for_geff(geff_path):
        if overwrite:
            delete_geff(geff_path, zarr_format=zarr_format)
        else:
            raise FileExistsError(
                "Found an existing geff present in `geff_path`. "
                "Please use `overwrite=True` or provide an alternative "
                "`geff_path` to write to."
            )

    tracks: dict[int, list[int]] = {}

    edges = []
    node_props: dict[str, list[int | float]] = {
        "id": [],
        "tracklet_id": [],
        "t": [],
        "x": [],
        "y": [],
    }

    segm_array = None
    node_id = 0

    sorted_files = sorted(ctc_path.glob("*.tif"))
    n_1_padding: tuple[int, ...] = ()
    expand_dims: tuple[None, ...] | slice = slice(None)

    for t, filepath in enumerate(sorted_files):
        frame = tifffile.imread(filepath)

        if segmentation_store is not None and segm_array is None:
            if frame.ndim == 3:
                node_props["z"] = []

            # created in first iteration
            if tczyx:
                n_1_padding = (1,) * (5 - frame.ndim - 1)  # forcing data to be (T, C, Z, Y, X)
                expand_dims = (np.newaxis,) * len(n_1_padding)

            # Warning is triggered when the zarr_format argument is ignored by zarr 2
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message=r".*ignoring keyword argument.*zarr_format.*",
                    category=UserWarning,
                )
                segm_array = zarr.open_array(
                    segmentation_store,
                    shape=(len(sorted_files), *n_1_padding, *frame.shape),
                    chunks=(1, *n_1_padding, *frame.shape),
                    dtype=frame.dtype,
                    mode="w" if overwrite else "w-",
                    zarr_format=zarr_format,
                )

        if segm_array is not None:
            segm_array[t] = frame[expand_dims]

        for obj in regionprops(frame):
            tracklet_id = obj.label
            node_props["id"].append(node_id)
            node_props["tracklet_id"].append(tracklet_id)
            node_props["t"].append(t)
            # using y,x for 2d and z,y,x for 3d
            for c, v in zip(("x", "y", "z"), obj.centroid[::-1], strict=False):
                node_props[c].append(v)

            if tracklet_id not in tracks:
                tracks[tracklet_id] = []

            tracks[tracklet_id].append(node_id)
            node_id += 1

    if len(node_props["id"]) == 0:
        raise ValueError(f"No nodes found in the CTC directory {ctc_path}")

    for _node_ids in tracks.values():
        # connect simple-paths of each track
        for i in range(len(_node_ids) - 1):
            # forward in time (parent -> child)
            edges.append((_node_ids[i], _node_ids[i + 1]))

    tracks_table = np.loadtxt(tracks_file_path, dtype=int)

    # removing orphan tracklets
    tracks_table = tracks_table[tracks_table[:, -1] > 0]

    for row in tracks_table:
        child_track_id = row[0]
        parent_track_id = row[-1]
        child_node_id = tracks[child_track_id][0]
        parent_node_id = tracks[parent_track_id][-1]
        # forward in time (parent -> child)
        edges.append((parent_node_id, child_node_id))

    axis_names = [
        Axis(name="t", type="time"),
        Axis(name="y", type="space"),
        Axis(name="x", type="space"),
    ]
    if "z" in node_props:
        axis_names.insert(1, Axis(name="z", type="space"))

    node_ids = np.asarray(node_props.pop("id"), dtype="uint")

    rel_objs = None
    if segmentation_store is not None:
        # Record related object metadata for segmentation
        seg_path = None
        if isinstance(segmentation_store, Path) or isinstance(segmentation_store, str):
            seg_path = segmentation_store
        else:
            try:
                if zarr.__version__.startswith("2"):
                    seg_path = segmentation_store.path  # type: ignore
                else:
                    seg_path = segmentation_store.root  # type: ignore
            except AttributeError:
                warnings.warn(
                    "Cannot determine path to segmentation_store for related objects metadata",
                    stacklevel=2,
                )

        if seg_path is not None:
            rel_path = os.path.relpath(seg_path, geff_path)
            rel_objs = [RelatedObject(type="labels", path=rel_path, label_prop="tracklet_id")]

    write_arrays(
        geff_store=geff_path,
        node_ids=node_ids,
        node_props={
            name: {"values": np.asarray(values), "missing": None}
            for name, values in node_props.items()
        },
        edge_ids=np.asarray(edges, dtype=node_ids.dtype),
        edge_props={},
        metadata=GeffMetadata(
            geff_version=geff.__version__,
            axes=axis_names,
            directed=True,
            node_props_metadata={},
            edge_props_metadata={},
            track_node_props={"tracklet": "tracklet_id"},
            related_objects=rel_objs,
        ),
        zarr_format=zarr_format,
    )



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from collections.abc import Sequence

    from numpy.typing import ArrayLike

    from geff._typing import InMemoryGeff
    from geff.metadata._schema import GeffMetadata


def has_valid_seg_id(memory_geff: InMemoryGeff, seg_id: str = "seg_id") -> tuple[bool, list[str]]:
    """
    Validate that all nodes in the geff have a property 'seg_id', that is of type integer.

    Args:
        memory_geff (InMemoryGeff): An InMemoryGeff to check for seg_id
        seg_id (str): the key to the dataset storing the segmentation value.

    Returns:
        tuple (bool, list[str])
        True if the checks passed, False if validation failed due to missing data or
        non-integer seg_ids.
        A list of encountered errors.
    """

    errors: list[str] = []

    # check that 'seg_id' property is present in the nodes group
    if seg_id not in memory_geff["node_props"]:
        errors.append("Missing seg_id property in Zarr store")
        return False, errors

    seg_ids = memory_geff["node_props"][seg_id]["values"]
    # Check that seg_id array is of integer type
    if not np.issubdtype(seg_ids.dtype, np.integer):
        errors.append(f"'seg_id' array has non-integer dtype: {seg_ids.dtype}")
        return False, errors

    # Check for no missing seg ids
    missing = memory_geff["node_props"][seg_id]["missing"]
    if missing is not None and any(missing):
        errors.append("Mismatch in number of node IDs and seg_ids.")
        return False, errors

    return len(errors) == 0, errors


def axes_match_seg_dims(
    memory_geff: InMemoryGeff,
    segmentation: ArrayLike,
) -> tuple[bool, list[str]]:
    """Validate that geff axes metadata have the same number of dimensions as the
    segmentation data.

    Args:
        memory_geff (InMemoryGeff): An InMemoryGeff to validate
        segmentation (ArrayLike): a 3D or 4D segmentation array (t, (z), y, x).

    Returns:
        tuple (bool, list[str])
        True if all checks passed
        False if the store does not provide metadata or if the number of dimensions in the
        metadata does not match that of the segmentation.
        A list of length 0 or 1 with the encountered error, if any.
    """

    errors: list[str] = []
    axes = memory_geff["metadata"].axes

    if axes:
        return np.asanyarray(segmentation).ndim == len(axes), errors
    else:
        errors.append("No axes metadata found in this geff.")
        return False, errors


def graph_is_in_seg_bounds(
    memory_geff: InMemoryGeff,
    segmentation: ArrayLike,
    scale: Sequence[float] | None = None,
) -> tuple[bool, list[str]]:
    """Validate that geff axes metadata have the same number of dimensions as the
    segmentation data.

    Args:
        memory_geff (InMemoryGeff): An InMemoryGeff to validate
        segmentation (ArrayLike): a 3D or 4D segmentation array (t, (z), y, x).
        scale (tuple[float] | list[float] | None = None): optional scaling tuple, with the
            same length as the number of dimensions in the segmentation data.

    Returns:
        tuple (bool, list[str])
        True if all checks passed
        False if the store does not provide metadata, if the provided scale tuple does not
            have the same length as the number of dimensions in the segmentation, if the
            number of dimensions in the metadata does not match that of the segmentation,
            or if the (scaled) graph data is not within the segmentation bounds.
        A list of length 0 or 1 with the encountered error, if any.
    """

    errors: list[str] = []
    axes = memory_geff["metadata"].axes

    segmentation = np.asanyarray(segmentation)
    seg_shape = segmentation.shape

    if scale is None:
        scale = [1.0] * segmentation.ndim

    if len(scale) != segmentation.ndim:
        errors.append(
            f"Length of scale factor list ({len(scale)} does not match with the number "
            f"of dimensions in the segmentation ({segmentation.ndim})"
        )
        return False, errors

    if axes:
        for i, ax in enumerate(axes):
            max_bound = ax.max
            if max_bound:
                if seg_shape[i] * scale[i] <= max_bound:
                    errors.append(
                        f"Graph axis {i} is out of bounds with value {max_bound} in "
                        f"segmentation axis size {seg_shape[i]} and scale factor "
                        f"{scale[i]}"
                    )
                    return False, errors
            else:
                errors.append("No axis 'max' value found in this geff metadata.")
                return False, errors
        return True, errors
    else:
        errors.append("No axes metadata found in this geff.")
        return False, errors


def has_seg_ids_at_time_points(
    segmentation: ArrayLike,
    time_points: Sequence[int],
    seg_ids: Sequence[int],
    metadata: GeffMetadata | None = None,
) -> tuple[bool, list[str]]:
    """
    Validates that labels with given seg_ids exist at time points t. If a store is
    provided, the time axis will be identified by the metadata using the 'type' key. If
    this is not possible, it is assumed that time is on axis 0.

    Args:
        segmentation (ArrayLike): a 3D or 4D segmentation array (t, (z), y, x).
        time_points (Sequence[int]): Sequence of time points to check.
        seg_ids (Sequence[int]): Sequence of seg_ids to check.
        metadata (GeffMetadata): If provided, it will attempt to read the axis order from the
            metadata. Otherwise, it is assumed that the dimension order is t(z)yx.

    Returns:
        tuple (bool, list[str])
        True if all seg_ids are present at their respective time points, False if an
        Index error is encountered or if there are any missing labels.
        A list of encountered errors (can be of length 1 when returning early).
    """

    errors: list[str] = []
    time_index = 0
    if metadata is not None:
        # load the axes metadata to extract the axes order. If it is not present, assume
        # that time is the first axes.
        axes = metadata.axes

        # check the metadata to see if an alternative time index is provided there.
        if axes:
            time_indices = [axes.index(ax) for ax in axes if ax.type == "time"]
            if len(time_indices) == 1:
                time_index = time_indices[0]

    # Create dictionary to map multiple seg_ids to the same time point.
    seg_id_group = defaultdict(list)
    for t, seg_id in zip(time_points, seg_ids, strict=False):
        seg_id_group[t].append(seg_id)

    # Loop over all time points, collect the label values, and check if the seg_ids are
    # present
    missing = defaultdict(list)
    for t in time_points:
        try:
            labels = np.unique(np.take(segmentation, indices=t, axis=time_index))
        except IndexError as e:
            errors.append(f"Time point {t} is out of bounds: {e}")
            return False, errors

        label_set = set(labels.tolist())
        for seg_id in seg_id_group[t]:
            if seg_id not in label_set:
                errors.append(f"Missing seg_id {seg_id} at time {t}")
                missing[t].append(seg_id)

    if len(missing) > 0:
        return False, errors
    return True, errors


def has_seg_ids_at_coords(
    segmentation: ArrayLike,
    coords: Sequence[Sequence[int]],
    seg_ids: Sequence[int],
    scale: Sequence[float] | None = None,
) -> tuple[bool, list[str]]:
    """
    Validates that the pixels at given coordinates in the segmentation have a value equal
      to the provided seg_ids.

    Args:
        segmentation (ArrayLike): a 3D or 4D segmentation array (t, (z), y, x).
        coords (Sequence[Sequence[int]]): Sequence of t(z)yx coordinates, should have the
            same order as the segmentation dimensions.
        seg_ids (Sequence[int]): Sequence of corresponding seg_ids to check.
        scale (Sequence[float] | None = None): optional scaling tuple, with the same
            length as the number of dimensions in the segmentation data.

    Returns:
        tuple[bool, list[str]]:
        True if all checks pass, False if an error was encountered or if there is no
        match.
        A list of encountered errors (can be of length 1 when returning early).

    """

    errors: list[str] = []
    if not len(coords) == len(seg_ids):
        errors.append("Coordinate list must have the same length as the list of seg_ids to test.")
        return False, errors

    segmentation = np.asanyarray(segmentation)
    if scale is None:
        scale = [1.0] * segmentation.ndim

    if len(scale) != segmentation.ndim:
        errors.append(
            f"Length of scale factor list ({len(scale)} does not match with the number "
            f"of dimensions in the segmentation ({segmentation.ndim})"
        )
        return False, errors

    missing = {}
    for coord, seg_id in zip(coords, seg_ids, strict=False):
        try:
            scaled_coord = [int(c * s) for c, s in zip(coord, scale, strict=True)]
            value = segmentation[tuple(scaled_coord)]
        except IndexError:
            errors.append(
                f"Coords {coord} are out of bounds for segmentation data with shape"
                f"{segmentation.shape} and scale factors {scale}"
            )
            return False, errors

        if value != seg_id:
            missing[seg_id] = coords

    if len(missing) > 0:
        return False, errors
    return True, errors



from collections.abc import Sequence
from typing import Any, Protocol, TypeVar

from geff_spec import GeffMetadata

T = TypeVar("T", covariant=True)


class GraphAdapter(Protocol[T]):
    def __init__(self, graph: T) -> None: ...

    def get_node_ids(self) -> Sequence[int]:
        """
        Get the node ids of the graph.

        Returns:
            node_ids (Sequence[int]): The node ids.
        """
        ...

    def get_edge_ids(self) -> Sequence[tuple[int, int]]:
        """
        Get the edges of the graph.

        Returns:
            edge_ids (Sequence[tuple[int, int]]): Pairs of node ids that represent edges.
        """
        ...

    def has_node_prop(self, name: str, node: int, metadata: GeffMetadata) -> bool:
        """Determine if the given node has a property value or not

        Args:
            name (str): The name of the node property.
            node (int): A node id to get the property for
            metadata (GeffMetadata): The GEFF metadata.

        Returns:
            bool: True if the node has the property, and False otherwise
        """
        ...

    def get_node_prop(self, name: str, node: int, metadata: GeffMetadata) -> Any:
        """
        Get a property of a specific node

        Args:
            name (str): The name of the node property.
            node (int): A node id to get the property for
            metadata (GeffMetadata): The GEFF metadata.

        Returns:
            Any: The value of the selected property for the given node
        """
        ...

    def has_edge_prop(
        self,
        name: str,
        edge: tuple[int, int],
        metadata: GeffMetadata,
    ) -> bool:
        """
        Determine if an edge has a specific property

        Args:
            name (str): The name of the edge property.
            edge (tuple[int,int]): A tuple of node ids representing an edge.
            metadata (GeffMetadata): The GEFF metadata.

        Returns:
            bool: True if the edge does have the property, false otherwise.
        """
        ...

    def get_edge_prop(
        self,
        name: str,
        edge: tuple[int, int],
        metadata: GeffMetadata,
    ) -> Any:
        """
        Get a property of a specific edge

        Args:
            name (str): The name of the edge property.
            edge (tuple[int, int]): A tuple of node ids representing an edge
            metadata (GeffMetadata): The GEFF metadata.

        Returns:
            Any: The value of the selected property for a specific edge.
        """
        ...

    # TODO: add get roi?



from typing import TYPE_CHECKING, Any

from geff.convert._trackmate_xml import from_trackmate_xml_to_geff

if TYPE_CHECKING:
    from ._ctc import ctc_tiffs_to_zarr, from_ctc_to_geff

__all__ = [
    "ctc_tiffs_to_zarr",
    "from_ctc_to_geff",
    "from_trackmate_xml_to_geff",
    "geff_to_csv",
    "geff_to_dataframes",
]


def __getattr__(name: str) -> Any:
    if name == "ctc_tiffs_to_zarr":
        from geff.convert._ctc import ctc_tiffs_to_zarr

        return ctc_tiffs_to_zarr
    if name == "from_ctc_to_geff":
        from geff.convert._ctc import from_ctc_to_geff

        return from_ctc_to_geff
    if name == "geff_to_csv":
        from geff.convert._dataframe import geff_to_csv

        return geff_to_csv
    if name == "geff_to_dataframes":
        from geff.convert._dataframe import geff_to_dataframes

        return geff_to_dataframes



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from geff_spec import Axis


def validate_ellipsoid(covariance: np.ndarray, axes: list[Axis] | None) -> None:
    """Validate that ellipsoid data has a valid covariance matrix

    The first axis of the covariance array corresponds to the number of nodes. The
    remaining axes correspond to the number of spatial axes.

    Args:
        covariance (np.ndarray): Covariance array stored as values for an ellipsoid property
        axes (list[Axis]): List of Axis metadata

    Raises:
        ValueError: Must define space axes in order to have ellipsoid data
        ValueError: Ellipsoid covariance matrix must have 1 + number of spatial dimensions
        ValueError: Spatial dimensions of covariance matrix must be equal
        ValueError: Ellipsoid covariance matrices must be symmetric
        ValueError: Ellipsoid covariance matrices must be positive-definite
    """
    bad_axes = True
    spatial_dim = 0
    # Axes need to exist and contain spatial data
    if axes is not None:
        for ax in axes:
            if ax.type == "space":
                spatial_dim += 1

        if spatial_dim > 0:
            bad_axes = False

    if bad_axes:
        raise ValueError("Must define space axes in order to have ellipsoid data")

    if covariance.ndim != (exp_dim := spatial_dim + 1):
        raise ValueError(
            f"Ellipsoid covariance matrix must have {exp_dim} dimensions, got {covariance.ndim}"
        )

    if covariance.shape[1] != covariance.shape[2]:
        raise ValueError(
            f"Spatial dimensions of covariance matrix must be equal, got {covariance.shape[1:]}"
        )

    transpose = [0, *list(range(covariance.ndim - 1, 0, -1))]
    if not np.allclose(covariance, np.transpose(covariance, axes=transpose)):
        raise ValueError("Ellipsoid covariance matrices must be symmetric")

    if not np.all(np.linalg.eigvals(covariance) > 0):
        raise ValueError("Ellipsoid covariance matrices must be positive-definite")


def validate_sphere(radius: np.ndarray) -> None:
    """Validate that sphere data has nonzero radii and is 1d

    Args:
        radius (np.ndarray): Values array of a sphere property

    Raises:
        ValueError: Sphere radius values must be non-negative
        ValueError: Sphere radius values must be 1D
    """
    if radius.ndim != 1:
        raise ValueError(f"Sphere radius values must be 1D, got {radius.ndim} dimensions")

    if np.any(radius < 0):
        raise ValueError("Sphere radius values must be non-negative.")



import io
import re
from copy import deepcopy
from pathlib import Path

import networkx as nx
import pytest

import geff.convert._trackmate_xml as tm_xml
from geff import validate_structure
from geff._graph_libs._networkx import NxBackend
from geff.testing._utils import nx_is_equal
from geff_spec import PropMetadata

try:
    from lxml import etree as ET
except ImportError:
    import xml.etree.ElementTree as ET

TEST_DATA = Path(__file__).parent.parent / "data"


def test_preliminary_checks(tmp_path: Path) -> None:
    xml_path = Path("path/to/xml")
    geff_path = Path("path/to/geff")

    # Normal case
    tm_xml._preliminary_checks(tmp_path, geff_path, True)

    # FileNotFoundError for missing XML file
    with pytest.raises(FileNotFoundError):
        tm_xml._preliminary_checks(xml_path, geff_path, False)

    # FileExistsError for existing GEFF file
    with pytest.raises(FileExistsError):
        tm_xml._preliminary_checks(tmp_path, tmp_path, False)


def test_get_units() -> None:
    space_warning = "No space unit found in the XML file. Setting to 'pixel'."
    time_warning = "No time unit found in the XML file. Setting to 'frame'."

    # Both spatial and time units
    xml_data = """<Model spatialunits="µm" timeunits="min"></Model>"""
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    obtained = tm_xml._get_units(element)
    expected = {"spatialunits": "µm", "timeunits": "min"}
    assert obtained == expected

    # Missing spatial units
    xml_data = """<Model timeunits="min"></Model>"""
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    with pytest.warns(UserWarning, match=space_warning):
        obtained = tm_xml._get_units(element)
    expected = {"spatialunits": "pixel", "timeunits": "min"}
    assert obtained == expected

    # Missing time units
    xml_data = """<Model spatialunits="µm"></Model>"""
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    with pytest.warns(UserWarning, match=time_warning):
        obtained = tm_xml._get_units(element)
    expected = {"spatialunits": "µm", "timeunits": "frame"}
    assert obtained == expected

    # Missing both spatial and time units
    xml_data = """<Model></Model>"""
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    with pytest.warns() as warning_list:
        obtained = tm_xml._get_units(element)
    expected = {"spatialunits": "pixel", "timeunits": "frame"}
    assert obtained == expected
    assert len(warning_list) == 2
    assert space_warning in str(warning_list[0].message)
    assert time_warning in str(warning_list[1].message)


def test_get_attributes_metadata() -> None:
    # Several attributes with Feature tags
    xml_data = """
        <FeatureDeclarations>
            <SpotFeatures>
                <Feature feature="QUALITY" isint="false" />
                <Feature feature="FRAME" isint="true" />
            </SpotFeatures>
        </FeatureDeclarations>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    obtained_attrs = tm_xml._get_attributes_metadata(it, element)
    expected_attrs = {
        "QUALITY": {"isint": "false"},
        "FRAME": {"isint": "true"},
    }
    assert obtained_attrs == expected_attrs

    # Without any Feature tags
    xml_data = """<SpotFeatures></SpotFeatures>"""
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    obtained_attrs = tm_xml._get_attributes_metadata(it, element)
    assert obtained_attrs == {}

    # With non Feature tag
    xml_data = """
        <FeatureDeclarations>
            <SpotFeatures>
                <Feature feature="QUALITY" isint="false" />
                <Other feature="FRAME" isint="true" />
            </SpotFeatures>
        </FeatureDeclarations>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    obtained_attrs = tm_xml._get_attributes_metadata(it, element)
    expected_attrs = {"QUALITY": {"isint": "false"}}
    assert obtained_attrs == expected_attrs


def test_convert_attributes() -> None:
    # Normal conversion with various data types
    attrs_md = {
        "feat_float": {"name": "feat_float", "isint": "false", "random": "info1"},
        "feat_int": {"name": "feat_int", "isint": "true", "random": "info1"},
        "feat_neg": {"name": "feat_neg", "isint": "true", "random": "info2"},
        "feat_string": {"name": "feat_string", "isint": "false", "random": "info3"},
    }
    converted_attrs = {
        "feat_float": "30",
        "feat_int": "20",
        "feat_neg": "-10",
        "feat_string": "nope",
    }
    tm_xml._convert_attributes(converted_attrs, attrs_md, "node", 1)
    expected_attr = {
        "feat_float": 30.0,
        "feat_int": 20,
        "feat_neg": -10.0,
        "feat_string": "nope",
    }
    assert converted_attrs == expected_attr

    # Special attributes
    attrs_md = {}
    converted_attrs = {"ID": "42", "name": "ID42", "ROI_N_POINTS": "10"}
    tm_xml._convert_attributes(converted_attrs, attrs_md, "node", 1)
    expected_attr = {"ID": 42, "name": "ID42", "ROI_N_POINTS": 10}
    assert converted_attrs == expected_attr

    # ValueError for invalid integer conversion
    attrs_md = {
        "feat_int": {"name": "feat_int", "isint": "true", "random": "info1"},
    }
    converted_attrs = {"feat_int": "not_an_int"}
    with pytest.raises(ValueError, match="Invalid integer value for feat_int: not_an_int"):
        tm_xml._convert_attributes(converted_attrs, attrs_md, "node", 1)

    # Missing attribute in metadata
    attrs_md = {
        "feat_float": {"name": "feat_float", "isint": "false", "random": "info1"},
        "feat_string": {"name": "feat_string", "isint": "false", "random": "info3"},
    }
    converted_attrs = {"feat_int": "10"}
    with pytest.warns(
        UserWarning, match="Node attribute feat_int not found in the attributes metadata."
    ):
        tm_xml._convert_attributes(converted_attrs, attrs_md, "node", 1)


def test_convert_ROI_coordinates() -> None:
    # 2D points
    el_obtained = ET.Element("Spot")
    el_obtained.attrib["ROI_N_POINTS"] = "3"
    el_obtained.text = "1 2.0 -3 -4.0 5.5 6"
    attr_obtained = deepcopy(el_obtained.attrib)
    attr_obtained["ROI_N_POINTS"] = int(attr_obtained["ROI_N_POINTS"])
    tm_xml._convert_ROI_coordinates(el_obtained, attr_obtained)
    attr_expected = {
        "ROI_N_POINTS": 3,
        "ROI_coords": [(1.0, 2.0), (-3.0, -4.0), (5.5, 6.0)],
    }
    assert attr_obtained == attr_expected

    # 3D points
    el_obtained = ET.Element("Spot")
    el_obtained.attrib["ROI_N_POINTS"] = "2"
    el_obtained.text = "1 2.0 -3 -4.0 5.5 6"
    attr_obtained = deepcopy(el_obtained.attrib)
    attr_obtained["ROI_N_POINTS"] = int(attr_obtained["ROI_N_POINTS"])
    tm_xml._convert_ROI_coordinates(el_obtained, attr_obtained)
    attr_expected = {
        "ROI_N_POINTS": 2,
        "ROI_coords": [(1.0, 2.0, -3.0), (-4.0, 5.5, 6.0)],
    }
    assert attr_obtained == attr_expected

    # KeyError for missing ROI_N_POINTS
    el_obtained = ET.Element("Spot")
    el_obtained.text = "1 2.0 -3 -4.0 5.5 6"
    attr_obtained = deepcopy(el_obtained.attrib)
    with pytest.raises(
        KeyError, match="No key 'ROI_N_POINTS' in the attributes of current element 'Spot'"
    ):
        tm_xml._convert_ROI_coordinates(el_obtained, attr_obtained)

    # TypeError for invalid ROI_N_POINTS
    el_obtained = ET.Element("Spot")
    el_obtained.text = "1 2.0 -3 -4.0 5.5 6"
    el_obtained.attrib["ROI_N_POINTS"] = "not_an_int"
    attr_obtained = deepcopy(el_obtained.attrib)
    with pytest.raises(TypeError, match="ROI_N_POINTS should be an integer"):
        tm_xml._convert_ROI_coordinates(el_obtained, attr_obtained)

    # No coordinates
    el_obtained = ET.Element("Spot")
    el_obtained.attrib["ROI_N_POINTS"] = "2"
    attr_obtained = deepcopy(el_obtained.attrib)
    attr_obtained["ROI_N_POINTS"] = int(attr_obtained["ROI_N_POINTS"])
    tm_xml._convert_ROI_coordinates(el_obtained, attr_obtained)
    attr_expected = {"ROI_N_POINTS": 2, "ROI_coords": None}
    assert attr_obtained == attr_expected


def test_add_all_nodes() -> None:
    # Several attributes
    xml_data = """
        <data>
           <frame>
               <Spot name="ID1000" ID="1000" x="10" y="20" />
               <Spot name="ID1001" ID="1001" x="30.5" y="30" />
           </frame>
        </data>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    attrs_md = {
        "x": {"name": "x", "isint": "false", "random": "info1"},
        "y": {"name": "y", "isint": "true", "random": "info3"},
    }
    obtained = nx.DiGraph()
    tm_xml._add_all_nodes(it, element, attrs_md, obtained)
    expected = nx.DiGraph()
    expected.add_nodes_from(
        [
            (1001, {"name": "ID1001", "y": 30, "ID": 1001, "x": 30.5}),
            (1000, {"name": "ID1000", "ID": 1000, "x": 10.0, "y": 20}),
        ]
    )
    assert nx_is_equal(obtained, expected)

    # Only ID attribute
    xml_data = """
        <data>
           <frame>
               <Spot ID="1000" />
               <Spot ID="1001" />
           </frame>
        </data>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    obtained = nx.DiGraph()
    tm_xml._add_all_nodes(it, element, {}, obtained)
    expected = nx.DiGraph()
    expected.add_nodes_from([(1001, {"ID": 1001}), (1000, {"ID": 1000})])
    assert nx_is_equal(obtained, expected)

    # No nodes
    xml_data = """
        <data>
            <frame />
        </data>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    obtained = nx.DiGraph()
    tm_xml._add_all_nodes(it, element, {}, obtained)
    assert nx_is_equal(obtained, nx.DiGraph())

    # No ID attribute
    xml_data = """
        <data>
            <frame>
                <Spot />
                <Spot ID="1001" />
            </frame>
        </data>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    msg = (
        "No key 'ID' in the attributes of current element 'Spot'. "
        "Not adding this node to the graph."
    )
    with pytest.warns(UserWarning, match=msg):
        tm_xml._add_all_nodes(it, element, {}, nx.DiGraph())


def test_add_edge() -> None:
    # Normal case with several attributes
    xml_data = """<data SPOT_SOURCE_ID="1" SPOT_TARGET_ID="2" x="20.5" y="25" />"""
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    track_id = 0
    attrs_md = {
        "x": {"name": "x", "isint": "false", "random": "info1"},
        "y": {"name": "y", "isint": "true", "random": "info3"},
        "SPOT_SOURCE_ID": {"name": "SPOT_SOURCE_ID", "isint": "true", "random": "info2"},
        "SPOT_TARGET_ID": {"name": "SPOT_TARGET_ID", "isint": "true", "random": "info4"},
    }
    obtained = nx.DiGraph()
    tm_xml._add_edge(element, attrs_md, obtained, track_id)
    expected = nx.DiGraph()
    expected.add_edge(1, 2, x=20.5, y=25, SPOT_SOURCE_ID=1, SPOT_TARGET_ID=2)
    expected.nodes[1]["TRACK_ID"] = track_id
    expected.nodes[2]["TRACK_ID"] = track_id
    assert nx_is_equal(obtained, expected)

    # No edge attributes
    xml_data = """<data SPOT_SOURCE_ID="1" SPOT_TARGET_ID="2" />"""
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    track_id = 0
    attrs_md = {
        "x": {"name": "x", "isint": "false", "random": "info1"},
        "y": {"name": "y", "isint": "true", "random": "info3"},
        "SPOT_SOURCE_ID": {"name": "SPOT_SOURCE_ID", "isint": "true", "random": "info2"},
        "SPOT_TARGET_ID": {"name": "SPOT_TARGET_ID", "isint": "true", "random": "info4"},
    }
    obtained = nx.DiGraph()
    tm_xml._add_edge(element, attrs_md, obtained, track_id)
    expected = nx.DiGraph()
    expected.add_edge(1, 2, SPOT_SOURCE_ID=1, SPOT_TARGET_ID=2)
    expected.nodes[1]["TRACK_ID"] = track_id
    expected.nodes[2]["TRACK_ID"] = track_id
    assert nx_is_equal(obtained, expected)

    # Missing SPOT_TARGET_ID
    xml_data = """<data SPOT_SOURCE_ID="1" x="20.5" y="25" />"""
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    attrs_md = {
        "x": {"name": "x", "isint": "false", "random": "info1"},
        "y": {"name": "y", "isint": "true", "random": "info3"},
        "SPOT_SOURCE_ID": {"name": "SPOT_SOURCE_ID", "isint": "true", "random": "info2"},
        "SPOT_TARGET_ID": {"name": "SPOT_TARGET_ID", "isint": "true", "random": "info4"},
    }
    with pytest.warns(
        UserWarning,
        match=(
            "No key 'SPOT_SOURCE_ID' or 'SPOT_TARGET_ID' in the attributes of "
            "current element 'data'. Not adding this edge to the graph."
        ),
    ):
        tm_xml._add_edge(element, attrs_md, nx.DiGraph(), track_id)

    # Inconsistent TRACK_ID
    xml_data = """<data SPOT_SOURCE_ID="1" SPOT_TARGET_ID="2" x="20.5" y="25" />"""
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    attrs_md = {
        "x": {"name": "x", "isint": "false", "random": "info1"},
        "y": {"name": "y", "isint": "true", "random": "info3"},
        "SPOT_SOURCE_ID": {"name": "SPOT_SOURCE_ID", "isint": "true", "random": "info2"},
        "SPOT_TARGET_ID": {"name": "SPOT_TARGET_ID", "isint": "true", "random": "info4"},
    }
    obtained = nx.DiGraph()
    obtained.add_nodes_from([(1, {"TRACK_ID": 1}), (2, {"TRACK_ID": 2})])
    with pytest.raises(
        AssertionError,
        match="Incoherent track ID for nodes 1 and 2",
    ):
        tm_xml._add_edge(element, attrs_md, obtained, 1)


def test_build_tracks() -> None:
    # Normal case with several attributes
    xml_data = """
        <data>
            <Track TRACK_ID="1" name="blob">
                <Edge SPOT_SOURCE_ID="11" SPOT_TARGET_ID="12" x="10.5" y="20" />
                <Edge SPOT_SOURCE_ID="12" SPOT_TARGET_ID="13" x="30" y="30" />
            </Track>
            <Track TRACK_ID="2" name="blub">
                <Edge SPOT_SOURCE_ID="21" SPOT_TARGET_ID="22" x="15.2" y="25" />
            </Track>
        </data>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    attrs_md = {
        "x": {"name": "x", "isint": "false", "random": "info1"},
        "y": {"name": "y", "isint": "true", "random": "info3"},
        "SPOT_SOURCE_ID": {"name": "SPOT_SOURCE_ID", "isint": "true", "random": "info2"},
        "SPOT_TARGET_ID": {"name": "SPOT_TARGET_ID", "isint": "true", "random": "info4"},
        "TRACK_ID": {"name": "TRACK_ID", "isint": "true", "random": "info5"},
    }
    obtained = nx.DiGraph()
    obtained_tracks_attrib = tm_xml._build_tracks(it, element, attrs_md, obtained)
    obtained_tracks_attrib = sorted(obtained_tracks_attrib, key=lambda d: d["TRACK_ID"])
    expected = nx.DiGraph()
    expected.add_edge(11, 12, SPOT_SOURCE_ID=11, SPOT_TARGET_ID=12, x=10.5, y=20)
    expected.add_edge(12, 13, SPOT_SOURCE_ID=12, SPOT_TARGET_ID=13, x=30.0, y=30)
    expected.add_edge(21, 22, SPOT_SOURCE_ID=21, SPOT_TARGET_ID=22, x=15.2, y=25)
    expected.add_nodes_from(
        [
            (11, {"TRACK_ID": 1}),
            (12, {"TRACK_ID": 1}),
            (13, {"TRACK_ID": 1}),
            (21, {"TRACK_ID": 2}),
            (22, {"TRACK_ID": 2}),
        ]
    )
    expected_tracks_attrib = [
        {"TRACK_ID": 2, "name": "blub"},
        {"TRACK_ID": 1, "name": "blob"},
    ]
    expected_tracks_attrib = sorted(expected_tracks_attrib, key=lambda d: d["TRACK_ID"])
    assert nx_is_equal(obtained, expected)
    assert obtained_tracks_attrib == expected_tracks_attrib

    # No edges in the tracks
    xml_data = """
        <data>
            <Track TRACK_ID="1" name="blob" />
            <Track TRACK_ID="2" name="blub" />
        </data>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    attrs_md = {
        "TRACK_ID": {"name": "TRACK_ID", "isint": "true", "random": "info5"},
    }
    obtained = nx.DiGraph()
    obtained_tracks_attrib = tm_xml._build_tracks(it, element, attrs_md, obtained)
    obtained_tracks_attrib = sorted(obtained_tracks_attrib, key=lambda d: d["TRACK_ID"])
    expected = nx.DiGraph()
    expected_tracks_attrib = [
        {"TRACK_ID": 2, "name": "blub"},
        {"TRACK_ID": 1, "name": "blob"},
    ]
    expected_tracks_attrib = sorted(expected_tracks_attrib, key=lambda d: d["TRACK_ID"])
    assert nx_is_equal(obtained, expected)
    assert obtained_tracks_attrib == expected_tracks_attrib

    # No node ID
    xml_data = """
        <data>
            <Track TRACK_ID="1" name="blob">
                <Edge x="10" y="20" />
                <Edge x="30" y="30" />
            </Track>
            <Track TRACK_ID="2" name="blub">
                <Edge x="15" y="25" />
            </Track>
        </data>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    attrs_md = {
        "x": {"name": "x", "isint": "false", "random": "info1"},
        "y": {"name": "y", "isint": "true", "random": "info3"},
        "TRACK_ID": {"name": "TRACK_ID", "isint": "true", "random": "info5"},
    }
    obtained = nx.DiGraph()
    with pytest.warns(
        UserWarning,
        match=(
            "No key 'SPOT_SOURCE_ID' or 'SPOT_TARGET_ID' in the attributes of "
            "current element 'Edge'. Not adding this edge to the graph."
        ),
    ):
        tm_xml._build_tracks(it, element, attrs_md, obtained)

    # No track ID
    xml_data = """
        <data>
            <Track name="blob">
                <Edge SPOT_SOURCE_ID="11" SPOT_TARGET_ID="12" x="10" y="20" />
                <Edge SPOT_SOURCE_ID="12" SPOT_TARGET_ID="13" x="30" y="30" />
            </Track>
            <Track name="blub">
                <Edge SPOT_SOURCE_ID="21" SPOT_TARGET_ID="22" x="15" y="25" />
            </Track>
        </data>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    attrs_md = {
        "x": {"name": "x", "isint": "false", "random": "info1"},
        "y": {"name": "y", "isint": "true", "random": "info3"},
        "SPOT_SOURCE_ID": {"name": "SPOT_SOURCE_ID", "isint": "true", "random": "info2"},
        "SPOT_TARGET_ID": {"name": "SPOT_TARGET_ID", "isint": "true", "random": "info4"},
    }
    with pytest.raises(
        KeyError,
        match=(
            re.escape(
                "No key 'TRACK_ID' in the attributes of current element 'Track'. "
                "Please check the XML file."
            )
        ),
    ):
        tm_xml._build_tracks(it, element, attrs_md, nx.DiGraph())


def test_get_filtered_tracks_ID() -> None:
    # Normal case with TRACK_ID attributes
    xml_data = """
        <data>
            <TrackID TRACK_ID="0" />
            <TrackID TRACK_ID="1" />
        </data>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    obtained_ID = tm_xml._get_filtered_tracks_ID(it, element)
    expected_ID = [0, 1]
    assert obtained_ID.sort() == expected_ID.sort()

    # No TRACK_ID
    xml_data = """
        <data>
            <TrackID />
            <TrackID />
        </data>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    with pytest.warns(
        UserWarning,
        match=(
            "No key 'TRACK_ID' in the attributes of current element 'TrackID'. Ignoring this track."
        ),
    ):
        tm_xml._get_filtered_tracks_ID(it, element)

    # No Track elements
    xml_data = """
        <data>
            <tag />
            <tag />
        </data>
    """
    it = ET.iterparse(io.BytesIO(xml_data.encode("utf-8")), events=["start", "end"])
    _, element = next(it)
    with pytest.warns(
        UserWarning,
        match=(
            "No key 'TRACK_ID' in the attributes of current element 'tag'. Ignoring this track."
        ),
    ):
        tm_xml._get_filtered_tracks_ID(it, element)


def test_get_trackmate_version(tmp_path: Path) -> None:
    # Normal case with version attribute
    xml_path = TEST_DATA / "FakeTracks.xml"
    obtained = tm_xml._get_trackmate_version(xml_path)
    assert obtained == "8.0.0-SNAPSHOT-f411154ed1a4b9de350bbfe91c230cf3ae7639a3"

    # Several attributes in TrackMate element
    xml_data = """<?xml version="1.0" encoding="UTF-8"?>
        <TrackMate attr_before="before" version="1" other_attr="value" and_another="test">
            <Log>Some log content</Log>
        </TrackMate>
    """
    xml_file = tmp_path / "with_multiple_attributes.xml"
    xml_file.write_text(xml_data, encoding="utf-8")
    obtained = tm_xml._get_trackmate_version(xml_file)
    assert obtained == "1"

    # No version attribute - TrackMate element without version
    xml_data = """<?xml version="1.0" encoding="UTF-8"?>
        <TrackMate>
            <Log>Some log content</Log>
        </TrackMate>
    """
    xml_file = tmp_path / "no_version.xml"
    xml_file.write_text(xml_data, encoding="utf-8")
    obtained = tm_xml._get_trackmate_version(xml_file)
    assert obtained == "unknown"

    # No TrackMate element at all
    xml_data = """<?xml version="1.0" encoding="UTF-8"?>
        <SomeOtherRoot>
            <Data>content</Data>
        </SomeOtherRoot>
    """
    xml_file = tmp_path / "no_trackmate.xml"
    xml_file.write_text(xml_data, encoding="utf-8")
    obtained = tm_xml._get_trackmate_version(xml_file)
    assert obtained == "unknown"


def test_get_specific_tags() -> None:
    # Normal case with several tags
    xml_path = TEST_DATA / "FakeTracks.xml"
    tag_names = [
        "FeaturePenalties",  # empty element
        "GUIState",  # simple element with attrib
        "TrackFilterCollection",  # nested elements with attribs
    ]
    obtained = tm_xml._get_specific_tags(xml_path, tag_names, 1)

    track_filter_collection = ET.Element("TrackFilterCollection")
    ET.SubElement(
        track_filter_collection,
        "Filter",
        attrib={"feature": "TRACK_START", "value": "77.9607843137255", "isabove": "false"},
    )
    ET.SubElement(
        track_filter_collection,
        "Filter",
        attrib={
            "feature": "TOTAL_DISTANCE_TRAVELED",
            "value": "20.75402586236767",
            "isabove": "true",
        },
    )
    expected = {
        "FeaturePenalties": ET.Element("FeaturePenalties"),
        "GUIState": ET.Element("GUIState", attrib={"state": "ConfigureViews"}),
        "TrackFilterCollection": track_filter_collection,
    }

    def normalize_xml_string(xml_str: str) -> str:
        """Remove extra whitespace between XML elements."""
        normalized = re.sub(r">\s+<", "><", xml_str.strip())
        return normalized

    assert obtained.keys() == expected.keys()
    for key in obtained:
        assert key in expected
        obtained_xml = normalize_xml_string(ET.tostring(obtained[key], encoding="unicode"))
        expected_xml = normalize_xml_string(ET.tostring(expected[key], encoding="unicode"))
        assert obtained_xml == expected_xml

    # Missing tags
    tag_names = ["FeaturePenalties", "MissingTag", "GUIState", "AnotherMissingTag"]
    with pytest.warns(
        UserWarning,
        match=r"Missing tag\(s\): \['MissingTag', 'AnotherMissingTag'\]",
    ):
        tm_xml._get_specific_tags(xml_path, tag_names, 1)


def test_extract_image_path() -> None:
    # Normal case with both filename and folder
    settings_element = ET.Element("Settings")
    ET.SubElement(
        settings_element,
        "ImageData",
        attrib={"filename": "test_image.tif", "folder": "/path/to/images"},
    )
    obtained = tm_xml._extract_image_path(settings_element)
    expected = str(Path("/path/to/images") / "test_image.tif")
    assert obtained == expected

    # Only filename, no folder
    settings_element = ET.Element("Settings")
    ET.SubElement(
        settings_element, "ImageData", attrib={"filename": "test_image.tif", "folder": ""}
    )
    with pytest.warns(
        UserWarning,
        match="No image folder found in the XML file",
    ):
        obtained = tm_xml._extract_image_path(settings_element)
    expected = str(Path("test_image.tif"))
    assert obtained == expected

    # Only folder, no filename
    settings_element = ET.Element("Settings")
    ET.SubElement(
        settings_element,
        "ImageData",
        attrib={"filename": "", "folder": "/path/to/images"},
    )
    with pytest.warns(
        UserWarning,
        match="No image name found in the XML file.",
    ):
        obtained = tm_xml._extract_image_path(settings_element)
    expected = str(Path("/path/to/images"))
    assert obtained == expected

    # Neither filename nor folder
    settings_element = ET.Element("Settings")
    ET.SubElement(settings_element, "ImageData", attrib={"filename": "", "folder": ""})
    with pytest.warns(
        UserWarning,
        match="No image path found in the XML file.",
    ):
        obtained = tm_xml._extract_image_path(settings_element)
    assert obtained is None

    # Missing ImageData element
    settings_element = ET.Element("Settings")
    with pytest.warns(
        UserWarning,
        match="No 'ImageData' tag found in the XML file.",
    ):
        obtained = tm_xml._extract_image_path(settings_element)
    assert obtained is None

    # Missing Settings element (None input)
    with pytest.warns(
        UserWarning,
        match="No 'Settings' tag found in the XML file.",
    ):
        obtained = tm_xml._extract_image_path(None)
    assert obtained is None


def test_get_feature_name() -> None:
    # Normal case
    feature_element = ET.Element("Feature", attrib={"feature": "QUALITY", "name": "Quality"})
    obtained = tm_xml._get_feature_name(feature_element, "QUALITY", "node")
    assert obtained == "Quality"

    # Empty element, node
    feature_element = ET.Element("Feature", attrib={"isint": "false"})
    with pytest.warns(
        UserWarning,
        match="Missing field 'name' in 'FeatureDeclarations' node tag.",
    ):
        obtained = tm_xml._get_feature_name(feature_element, "QUALITY", "node")
        assert obtained == "QUALITY"

    # Empty element, edge
    feature_element = ET.Element("Feature", attrib={"isint": "false"})
    with pytest.warns(
        UserWarning,
        match="Missing field 'name' in 'FeatureDeclarations' edge tag.",
    ):
        obtained = tm_xml._get_feature_name(feature_element, "QUALITY", "edge")
        assert obtained == "QUALITY"


def test_get_feature_dtype() -> None:
    # Normal case, is int
    feat_element = ET.Element("Feature", attrib={"feature": "QUALITY", "isint": "true"})
    obtained = tm_xml._get_feature_dtype(feat_element, "node")
    assert obtained == "int"

    # Normal case, is not int
    feat_element = ET.Element("Feature", attrib={"feature": "QUALITY", "isint": "false"})
    obtained = tm_xml._get_feature_dtype(feat_element, "node")
    assert obtained == "float"

    # Empty element, node
    feat_element = ET.Element("Feature", attrib={"dimension": "TIME"})
    with pytest.raises(
        ValueError,
        match=re.escape("Missing field 'isint' in 'FeatureDeclarations' node tag."),
    ):
        tm_xml._get_feature_dtype(feat_element, "node")

    # Empty element, edge
    feat_element = ET.Element("Feature", attrib={"dimension": "TIME"})
    with pytest.raises(
        ValueError,
        match=re.escape("Missing field 'isint' in 'FeatureDeclarations' edge tag."),
    ):
        tm_xml._get_feature_dtype(feat_element, "edge")


def test_get_feature_unit() -> None:
    units = {"spatialunits": "micrometer", "timeunits": "second"}

    # NONE dimension
    feature_element = ET.Element("Feature", attrib={"dimension": "NONE"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)
    assert obtained == "None"

    # QUALITY dimension
    feature_element = ET.Element("Feature", attrib={"dimension": "QUALITY"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)
    assert obtained == "None"

    # POSITION dimension
    feature_element = ET.Element("Feature", attrib={"dimension": "POSITION"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)
    assert obtained == "micrometer"

    # LENGTH dimension
    feature_element = ET.Element("Feature", attrib={"dimension": "LENGTH"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)
    assert obtained == "micrometer"

    # TIME dimension
    feature_element = ET.Element("Feature", attrib={"dimension": "TIME"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)
    assert obtained == "second"

    # VELOCITY dimension
    feature_element = ET.Element("Feature", attrib={"dimension": "VELOCITY"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)
    assert obtained == "micrometer / second"

    # AREA dimension
    feature_element = ET.Element("Feature", attrib={"dimension": "AREA"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)
    assert obtained == "micrometer^2"

    # ANGLE dimension
    feature_element = ET.Element("Feature", attrib={"dimension": "ANGLE"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)
    assert obtained == "radian"

    # RATE dimension
    feature_element = ET.Element("Feature", attrib={"dimension": "RATE"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)
    assert obtained == "1 / second"

    # ANGLE_RATE dimension
    feature_element = ET.Element("Feature", attrib={"dimension": "ANGLE_RATE"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)
    assert obtained == "radian / second"

    # Empty units dictionary
    empty_units = {}
    feature_element = ET.Element("Feature", attrib={"dimension": "POSITION"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", empty_units)
    assert obtained == "pixel"  # default spatial unit

    feature_element = ET.Element("Feature", attrib={"dimension": "TIME"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", empty_units)
    assert obtained == "frame"  # default time unit

    # Partial units dictionary
    partial_units = {"spatialunits": "micrometer"}
    feature_element = ET.Element("Feature", attrib={"dimension": "VELOCITY"})
    obtained = tm_xml._get_feature_unit(feature_element, "SpotFeatures", partial_units)
    assert obtained == "micrometer / frame"  # default time unit

    # Unknown dimension
    feature_element = ET.Element("Feature", attrib={"dimension": "UNKNOWN_DIM"})
    with pytest.raises(ValueError, match="Unknown dimension 'UNKNOWN_DIM'"):
        tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)

    # Missing dimension attribute
    feature_element = ET.Element("Feature", attrib={})
    with pytest.raises(ValueError, match="Unknown dimension 'None'"):
        tm_xml._get_feature_unit(feature_element, "SpotFeatures", units)


def test_process_feature_metadata() -> None:
    units = {"spatialunits": "pixel", "timeunits": "frame"}

    # Normal case
    obtained = {}
    feature_element = ET.Element(
        "Feature",
        attrib={"feature": "QUALITY", "name": "Quality", "isint": "false", "dimension": "QUALITY"},
    )
    tm_xml._process_feature_metadata(feature_element, obtained, "SpotFeatures", units)
    expected = {
        "QUALITY": {"identifier": "QUALITY", "name": "Quality", "dtype": "float", "unit": "None"}
    }
    assert obtained == expected

    # Missing 'feature' field
    obtained = {}
    feature_element = ET.Element(
        "Feature", attrib={"name": "Quality", "isint": "false", "dimension": "QUALITY"}
    )
    with pytest.raises(KeyError, match="Missing field 'feature' in 'FeatureDeclarations'"):
        tm_xml._process_feature_metadata(feature_element, obtained, "SpotFeatures", units)

    # Dict with existing data
    obtained = {"EXISTING_FEATURE": {"feature": "EXISTING_FEATURE"}}
    feature_element = ET.Element(
        "Feature",
        attrib={
            "feature": "NEW_FEATURE",
            "name": "New feature",
            "isint": "true",
            "dimension": "TIME",
        },
    )
    tm_xml._process_feature_metadata(feature_element, obtained, "EdgeFeatures", units)
    expected = {
        "EXISTING_FEATURE": {"feature": "EXISTING_FEATURE"},
        "NEW_FEATURE": {
            "identifier": "NEW_FEATURE",
            "name": "New feature",
            "dtype": "int",
            "unit": "frame",
        },
    }
    assert obtained == expected

    # Duplicate feature identifier
    obtained = {"QUALITY": {"name": "Quality", "isint": "false", "dimension": "QUALITY"}}
    feature_element = ET.Element(
        "Feature",
        attrib={"feature": "QUALITY", "name": "Quality", "isint": "false", "dimension": "QUALITY"},
    )
    with pytest.raises(ValueError, match="Duplicate feature identifier 'QUALITY' found"):
        tm_xml._process_feature_metadata(feature_element, obtained, "SpotFeatures", units)

    # Error propagation
    obtained = {}
    feature_element = ET.Element(
        "Feature",
        attrib={
            "feature": "TEST_FEATURE",
            "name": "Test",
            "dimension": "UNKNOWN_DIMENSION",
            "isint": "false",
        },
    )
    with pytest.raises(ValueError, match="Unknown dimension 'UNKNOWN_DIMENSION'"):
        tm_xml._process_feature_metadata(feature_element, obtained, "SpotFeatures", units)


def test_metadata_info_stability(tmp_path: Path) -> None:
    """Test that metadata information remains stable after writing to GEFF."""
    # Metadata before writing to GEFF.
    xml_path = TEST_DATA / "FakeTracks.xml"
    units = {"spatialunits": "micrometer", "timeunits": "second"}
    segmentation = True
    props_md = tm_xml._extract_props_metadata(xml_path, units, segmentation)
    md_before_write = tm_xml._build_geff_metadata(
        xml_path=xml_path,
        units=units,
        img_path="/path/to/image.tif",
        trackmate_metadata={},
        props_metadata=props_md,
    )

    # Metadata after writing to GEFF and reading it back.
    geff_output = tmp_path / "test.geff"
    with pytest.warns():
        tm_xml.from_trackmate_xml_to_geff(TEST_DATA / "FakeTracks.xml", geff_output)
    _, md_after_write = NxBackend.read(geff_output, structure_validation=False)

    # Compare metadata before and after writing.
    assert (
        md_before_write.node_props_metadata["AREA"].unit
        == md_after_write.node_props_metadata["AREA"].unit
    ), "Metadata unit for 'AREA' does not match after writing."
    assert (
        md_before_write.node_props_metadata["AREA"] == md_after_write.node_props_metadata["AREA"]
    ), "Metadata for 'AREA' does not match after writing."

    # Check equality of props that are present before and after writing
    # A few props are expected to be dropped (SHAPE_INDEX, MANUAL_SPOT_COLOR)
    # and added (ID, TRACK_ID, name)
    shared_keys = set(md_before_write.node_props_metadata.keys()).intersection(
        set(md_after_write.node_props_metadata.keys())
    )
    shared_md_before = {k: md_before_write.node_props_metadata[k] for k in shared_keys}
    shared_md_after = {k: md_after_write.node_props_metadata[k] for k in shared_keys}
    assert shared_md_before == shared_md_after, "Metadata does not match after writing."


def test_build_geff_metadata() -> None:
    """Test that the node and edge metadata are correctly built."""
    xml_path = TEST_DATA / "FakeTracks.xml"
    units = {"spatialunits": "micrometer", "timeunits": "second"}
    img_path = "/path/to/image.tif"
    trackmate_metadata = {}
    props_metadata = {
        "node_props_metadata": {
            "prop1": {
                "identifier": "prop1",
                "dtype": "float",
                "unit": "micrometer",
                "name": "prop 1",
                "description": "description 1",
            },
            "prop2": {
                "identifier": "prop2",
                "dtype": "int",
                "varlength": True,
                "unit": "micrometer",
                "name": "prop 2",
                "description": "description 2",
            },
        },
        "edge_props_metadata": {
            "prop3": {
                "identifier": "prop3",
                "dtype": "float",
                "unit": "second",
                "name": "prop 3",
                "description": "description 3",
            }
        },
        "lineage_props_metadata": {
            "prop4": {
                "identifier": "prop4",
                "dtype": "bool",
                "name": "prop 4",
                "description": "description 4",
            }
        },
    }
    prop_md_1 = PropMetadata(
        identifier="prop1",
        dtype="float",
        unit="micrometer",
        name="prop 1",
        description="description 1",
    )
    prop_md_2 = PropMetadata(
        identifier="prop2",
        dtype="int",
        varlength=True,
        unit="micrometer",
        name="prop 2",
        description="description 2",
    )
    prop_md_3 = PropMetadata(
        identifier="prop3", dtype="float", unit="second", name="prop 3", description="description 3"
    )

    obtained_md = tm_xml._build_geff_metadata(
        xml_path=xml_path,
        units=units,
        img_path=img_path,
        trackmate_metadata=trackmate_metadata,
        props_metadata=props_metadata,
    )

    assert obtained_md.node_props_metadata["prop1"] == prop_md_1
    assert obtained_md.node_props_metadata["prop2"] == prop_md_2
    assert obtained_md.edge_props_metadata["prop3"] == prop_md_3


def _validate_ROI_info(geff_path: Path) -> None:
    """
    Check if the GEFF file contains ROI information.

    ROI information includes the node attribute 'ROI_N_POINTS' and 'ROI_coords'
    as well as the associated metadata.
    """
    graph, md = NxBackend.read(geff_path, structure_validation=False)

    # Data check
    for _, data in graph.nodes(data=True):
        assert "ROI_N_POINTS" in data, "Missing 'ROI_N_POINTS' in node attributes."
        assert "ROI_coords" in data, "Missing 'ROI_coords' in node attributes."

    # Metadata check
    for prop in ["ROI_N_POINTS", "ROI_coords"]:
        assert prop in md.node_props_metadata, f"Missing '{prop}' in node properties metadata."


def test_from_trackmate_xml_to_geff(tmp_path: Path) -> None:
    # No arguments, should use default values
    geff_output = tmp_path / "test.geff"
    with pytest.warns() as warning_list:
        tm_xml.from_trackmate_xml_to_geff(TEST_DATA / "FakeTracks.xml", geff_output)
    warning_messages = [str(warning.message) for warning in warning_list]
    assert any("node properties were removed from the metadata" in msg for msg in warning_messages)
    assert any("edge property was removed from the metadata" in msg for msg in warning_messages)
    validate_structure(geff_output)
    _validate_ROI_info(geff_output)

    # Discard filtered spots and tracks
    with pytest.warns() as warning_list:
        tm_xml.from_trackmate_xml_to_geff(
            TEST_DATA / "FakeTracks.xml",
            geff_output,
            discard_filtered_spots=True,
            discard_filtered_tracks=True,
            overwrite=True,
        )
    warning_messages = [str(warning.message) for warning in warning_list]
    assert any("node properties were removed from the metadata" in msg for msg in warning_messages)
    assert any("edge property was removed from the metadata" in msg for msg in warning_messages)
    validate_structure(geff_output)
    _validate_ROI_info(geff_output)

    # Geff file already exists
    with pytest.raises(FileExistsError):
        tm_xml.from_trackmate_xml_to_geff(TEST_DATA / "FakeTracks.xml", geff_output)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import networkx as nx
import numpy as np

if TYPE_CHECKING:
    from numpy.typing import ArrayLike


def validate_tracklets(
    node_ids: ArrayLike, edge_ids: ArrayLike, tracklet_ids: ArrayLike
) -> tuple[bool, list[str]]:
    """
    Validates if each tracklet forms a single, cycle-free path using NetworkX
    for improved performance.

    Args:
        node_ids (ArrayLike): Sequence of node identifiers.
        edge_ids (ArrayLike): Sequence of edges as (source, target) node ID pairs.
            Edges must be between nodes in `node_ids`.
        tracklet_ids (ArrayLike): Sequence of tracklet IDs corresponding to each node.

    Returns:
        tuple[bool, list[str]]:
            - is_valid (bool): True if all tracklets are valid, otherwise False.
            - errors (list[str]): List of error messages for invalid tracklets.
    """
    errors = []

    nodes = np.asarray(node_ids, dtype=np.int64)
    edges = np.asarray(edge_ids, dtype=np.int64)
    tracklets = np.asarray(tracklet_ids, dtype=np.int64)

    # Group nodes by tracklet ID.
    tracklet_to_nodes: dict[int, list[int]] = {}
    for node, t_id in zip(nodes, tracklets, strict=False):
        tracklet_to_nodes.setdefault(t_id, []).append(node)

    # Build the graph.
    G: nx.DiGraph[int] = nx.DiGraph(tuple(edge) for edge in edges)
    # Ensure all nodes from node_ids are in the graph, even if isolated.
    G.add_nodes_from(nodes)

    # Validate each tracklet.
    for t_id, t_nodes in tracklet_to_nodes.items():
        # by definition, a tracklet
        if len(t_nodes) < 2:
            continue

        # Gets a subgraph for the current tracklet.
        S = cast("nx.DiGraph[int]", G.subgraph(t_nodes))

        # Check - no branches or merges (junctions).
        max_in_degree = max((d for _, d in S.in_degree), default=0)
        max_out_degree = max((d for _, d in S.out_degree), default=0)

        if max_in_degree > 1 or max_out_degree > 1:
            errors.append(f"Tracklet {t_id}: Invalid path structure (branch or merge detected).")
            continue

        # Check - No cycles.
        if not nx.is_directed_acyclic_graph(S):
            errors.append(f"Tracklet {t_id}: Cycle detected.")
            continue

        # Check - Fully connected.
        if not nx.is_weakly_connected(S):
            errors.append(f"Tracklet {t_id}: Not fully connected.")
            continue

        # Check - Tracklet is maximal linear segment.
        start_node = next(n for n, d in S.in_degree if d == 0)
        end_node = next(n for n, d in S.out_degree if d == 0)

        # Check if the path could be extended backward
        preds_in_G = list(G.predecessors(start_node))
        if len(preds_in_G) == 1:
            predecessor = preds_in_G[0]
            # If the predecessor is also part of a linear segment...
            # TODO: remove ignore, see issue 314
            if G.out_degree(predecessor) == 1:  # pyright: ignore
                errors.append(
                    f"Tracklet {t_id}: Not maximal. Path can extend backward to node {predecessor}."
                )
                continue

        # Check if the path could be extended forward
        succs_in_G = list(G.successors(end_node))
        if len(succs_in_G) == 1:
            successor = succs_in_G[0]
            # If the successor is also part of a linear segment...
            # TODO: remove ignore, see issue 314
            if G.in_degree(successor) == 1:  # pyright: ignore
                errors.append(
                    f"Tracklet {t_id}: Not maximal. Path can extend forward to node {successor}."
                )
                continue

    return not errors, errors


def validate_lineages(
    node_ids: ArrayLike, edge_ids: ArrayLike, lineage_ids: ArrayLike
) -> tuple[bool, list[str]]:
    """Validates if each lineage is a valid, isolated connected component.

    A lineage is considered valid if and only if the set of nodes belonging
    to it is identical to one of the graph's weakly connected components.
    This efficiently ensures both internal connectivity and external isolation.

    Args:
        node_ids: A sequence of unique node identifiers in the graph.
        edge_ids: A sequence of (source, target) pairs representing directed
            edges.
        lineage_ids: A sequence of lineage identifiers corresponding to each
            node in `node_ids`.

    Returns:
        A tuple containing:
            - is_valid (bool): True if all lineages are valid connected
              components, False otherwise.
            - errors (list[str]): A list of error messages for each invalid
              lineage.
    """
    ID_DTYPE = np.int64
    # Ensure consistent dtypes.
    nodes = np.asarray(node_ids, dtype=ID_DTYPE)
    edges = np.asarray(edge_ids, dtype=ID_DTYPE)
    lineages = np.asarray(lineage_ids, dtype=ID_DTYPE)

    errors: list[str] = []
    lineage_to_nodes: dict[np.int64, list[np.int64]] = {}

    for node, l_id in zip(nodes, lineages, strict=False):
        lineage_to_nodes.setdefault(l_id, []).append(node)

    # Build the graph.
    G = nx.DiGraph(tuple(edge) for edge in edges)
    G.add_nodes_from(nodes)

    # Find all weakly connected components.
    valid_components = {frozenset(component) for component in nx.weakly_connected_components(G)}

    # Check if each lineage's node set matches a valid component.
    for l_id, l_nodes in lineage_to_nodes.items():
        if not l_nodes:
            continue

        l_nodes_set = frozenset(l_nodes)

        if l_nodes_set not in valid_components:
            errors.append(f"Lineage {l_id}: Does not form a valid, isolated connected component.")

    return not errors, errors



import numpy as np
import pytest

import geff.validate.data
from geff.testing.data import create_mock_geff, create_simple_2d_geff
from geff.validate.data import ValidationConfig, validate_data


class Test_validate_data:
    store, memory_geff = create_simple_2d_geff()

    def test_valid_graph(self):
        # test valid
        validate_data(self.memory_geff, ValidationConfig(graph=True))

    def test_nodes_for_edges(self, monkeypatch):
        # error on validate_nodes_for_edges
        monkeypatch.setattr(
            geff.validate.data, "validate_nodes_for_edges", lambda node_ids, edge_ids: (False, [])
        )
        with pytest.raises(ValueError, match="Some edges are missing nodes"):
            validate_data(self.memory_geff, ValidationConfig(graph=True))

    def test_no_self_edges(self, monkeypatch):
        monkeypatch.setattr(
            geff.validate.data, "validate_no_self_edges", lambda edge_ids: (False, [])
        )
        with pytest.raises(ValueError, match="Self edges found in data"):
            validate_data(self.memory_geff, ValidationConfig(graph=True))

    def test_no_repeated_edges(self, monkeypatch):
        monkeypatch.setattr(
            geff.validate.data, "validate_no_repeated_edges", lambda edge_ids: (False, [])
        )
        with pytest.raises(ValueError, match="Repeated edges found in data"):
            validate_data(self.memory_geff, ValidationConfig(graph=True))

    def test_unique_node_ids(self, monkeypatch):
        monkeypatch.setattr(
            geff.validate.data, "validate_unique_node_ids", lambda node_ids: (False, [])
        )
        with pytest.raises(ValueError, match="Some node ids are not unique"):
            validate_data(self.memory_geff, ValidationConfig(graph=True))

    def test_sphere(self):
        # Invalid spheres are tested in test_shapes
        # Only need to test a valid case
        _, memory_geff = create_mock_geff(
            node_id_dtype="int",
            node_axis_dtypes={"position": "float64", "time": "float64"},
            directed=True,
            num_nodes=10,
            num_edges=10,
            extra_node_props={"radius": "int"},
            include_t=True,
            include_z=False,  # 2D only
            include_y=True,
            include_x=True,
        )
        # Add sphere metadata
        memory_geff["metadata"].sphere = "radius"

        validate_data(config=ValidationConfig(sphere=True), memory_geff=memory_geff)

    def test_ellipsoid(self):
        # Invalid ellipsoids are tested in test_shapes
        # Only need to test a valid case
        _, memory_geff = create_mock_geff(
            node_id_dtype="int",
            node_axis_dtypes={"position": "float64", "time": "float64"},
            directed=True,
            num_nodes=10,
            num_edges=10,
            extra_node_props={"covariance2d": "float64"},
            include_t=True,
            include_z=False,  # 2D only
            include_y=True,
            include_x=True,
        )
        # Add ellipsoid metadata
        memory_geff["metadata"].ellipsoid = "covariance2d"
        # Overwrite ellipsoid values
        covar = np.ones((10, 2, 2))
        covar[:, 0, 0] = 2
        covar[:, 1, 1] = 2
        memory_geff["node_props"]["covariance2d"]["values"] = covar

        validate_data(config=ValidationConfig(ellipsoid=True), memory_geff=memory_geff)

    def test_tracklet(self, monkeypatch):
        # validate_tracklets is tested in test_graph
        # Just need to trigger the value error
        _, memory_geff = create_mock_geff(
            node_id_dtype="int",
            node_axis_dtypes={"position": "float64", "time": "float64"},
            directed=True,
            num_nodes=10,
            num_edges=10,
            extra_node_props={"tracklet": "int"},
            include_t=True,
            include_z=False,  # 2D only
            include_y=True,
            include_x=True,
        )
        # Add tracklet metadata
        memory_geff["metadata"].track_node_props = {"tracklet": "tracklet"}

        # Monkeypatch validate tracklets to return false
        monkeypatch.setattr(geff.validate.data, "validate_tracklets", lambda x, y, z: (False, []))

        with pytest.raises(ValueError, match="Found invalid tracklets"):
            validate_data(config=ValidationConfig(tracklet=True), memory_geff=memory_geff)

    def test_lineages(self, monkeypatch):
        # validate_lineages is tested in test_graph
        # Just need to trigger the value error
        _, memory_geff = create_mock_geff(
            node_id_dtype="int",
            node_axis_dtypes={"position": "float64", "time": "float64"},
            directed=True,
            num_nodes=10,
            num_edges=10,
            extra_node_props={"lineage": "int"},
            include_t=True,
            include_z=False,  # 2D only
            include_y=True,
            include_x=True,
        )
        # Add tracklet metadata
        memory_geff["metadata"].track_node_props = {"lineage": "lineage"}

        # Monkeypatch validate tracklets to return false
        monkeypatch.setattr(geff.validate.data, "validate_lineages", lambda x, y, z: (False, []))

        with pytest.raises(ValueError, match="Found invalid lineages"):
            validate_data(config=ValidationConfig(lineage=True), memory_geff=memory_geff)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, TypeAlias, TypeVar, get_args, overload

from geff.core_io._utils import check_for_geff, delete_geff, remove_tilde

from ._errors import MissingDependencyError

R = TypeVar("R", covariant=True)

if TYPE_CHECKING:
    import networkx as nx
    import rustworkx as rx
    import spatial_graph as sg
    from numpy.typing import NDArray
    from zarr.storage import StoreLike

    from geff._typing import PropDictNpArray
    from geff.validate.data import ValidationConfig
    from geff_spec import AxisType, GeffMetadata

    from ._backend_protocol import Backend

    NxGraph: TypeAlias = nx.Graph[Any] | nx.DiGraph[Any]
    RxGraph: TypeAlias = rx.PyGraph | rx.PyDiGraph
    SgGraph: TypeAlias = sg.SpatialGraph | sg.SpatialDiGraph
    SupportedGraphType: TypeAlias = NxGraph | RxGraph | SgGraph

SupportedBackend = Literal["networkx", "rustworkx", "spatial-graph"]
"""Supported graph library backends for reading to and writing from."""

AVAILABLE_BACKENDS: list[Backend] = []


# This function will add available backends to the available backend list
# it has to be called after the definition of get_backends
def _import_available_backends() -> None:
    backends: tuple[SupportedBackend] = get_args(SupportedBackend)
    for backend in backends:
        try:
            AVAILABLE_BACKENDS.append(get_backend(backend))
        except MissingDependencyError:
            pass


# NOTE: overload get_backend for new backends by typing the return type as Backend[GraphType]
@overload
def get_backend(backend: Literal["networkx"]) -> Backend[NxGraph]: ...
@overload
def get_backend(backend: Literal["rustworkx"]) -> Backend[RxGraph]: ...
@overload
def get_backend(backend: Literal["spatial-graph"]) -> Backend[SgGraph]: ...
def get_backend(backend: SupportedBackend) -> Backend:
    """
    Get a specified backend io module.

    Args:
        backend (SupportedBackend): Flag for the chosen backend.

    Returns:
        Backend: A module for reading and writing GEFF data to and from the specified backend.
    """
    match backend:
        case "networkx":
            from geff._graph_libs._networkx import NxBackend

            return NxBackend()
        case "rustworkx":
            from geff._graph_libs._rustworkx import RxBackend

            return RxBackend()
        case "spatial-graph":
            from geff._graph_libs._spatial_graph import SgBackend

            return SgBackend()
        # Add cases for new backends, remember to add overloads
        case _:
            raise ValueError(f"Unsupported backend chosen: '{backend}'")


_import_available_backends()


# Used in the write function wrapper, where the backend should be determined from the graph type
def get_backend_from_graph_type(
    graph: SupportedGraphType,
) -> Backend[SupportedGraphType]:
    for backend_module in AVAILABLE_BACKENDS:
        if isinstance(graph, backend_module.GRAPH_TYPES):
            return backend_module
    raise TypeError(f"Unrecognized graph type '{type(graph)}'.")


# NOTE: when overloading read for a new backend, if additional arguments can be accepted, explicitly
# define them such as in the spatial-graph overload below, where position_attr has been added.
@overload
def read(
    store: StoreLike,
    structure_validation: bool = ...,
    node_props: list[str] | None = ...,
    edge_props: list[str] | None = ...,
    data_validation: ValidationConfig | None = ...,
    *,
    backend: Literal["networkx"] = "networkx",
) -> tuple[NxGraph, GeffMetadata]: ...
@overload
def read(
    store: StoreLike,
    structure_validation: bool = ...,
    node_props: list[str] | None = ...,
    edge_props: list[str] | None = ...,
    data_validation: ValidationConfig | None = ...,
    *,
    backend: Literal["rustworkx"],
) -> tuple[RxGraph, GeffMetadata]: ...
@overload
def read(
    store: StoreLike,
    structure_validation: bool = ...,
    node_props: list[str] | None = ...,
    edge_props: list[str] | None = ...,
    data_validation: ValidationConfig | None = ...,
    *,
    backend: Literal["spatial-graph"],
    position_attr: str = "position",
) -> tuple[SgGraph, GeffMetadata]: ...
def read(
    store: StoreLike,
    structure_validation: bool = True,
    node_props: list[str] | None = None,
    edge_props: list[str] | None = None,
    data_validation: ValidationConfig | None = None,
    *,
    backend: SupportedBackend = "networkx",
    **backend_kwargs: Any,
) -> tuple[SupportedGraphType, GeffMetadata]:
    """
    Read a GEFF to a chosen backend.

    Args:
        store (zarr.storage.StoreLike): The path or zarr store to the root of the geff zarr, where
            the .attrs contains the geff  metadata.
        structure_validation (bool, optional): Flag indicating whether to perform validation on the
            geff file before loading into memory. If set to False and there are
            format issues, will likely fail with a cryptic error. Defaults to True.
        node_props (list[str], optional): The names of the node properties to load,
            if None all properties will be loaded, defaults to None.
        edge_props (list[str], optional): The names of the edge properties to load,
            if None all properties will be loaded, defaults to None.
        backend (SupportedBackend): Flag for the chosen backend, default
            is "networkx".
        data_validation (ValidationConfig, optional): Optional configuration for which
            optional types of data to validate. Each option defaults to False.
        backend_kwargs (Any): Additional kwargs that may be accepted by
            the backend when reading the data.

    Returns:
        graph (tuple[networkx.Graph | networkx.DiGraph | rustworkx.PyGraph | rustworkx.PyDiGraph | spatial_graph.SpatialGraph | spatial_graph.SpatialDiGraph, GeffMetadata]):
            Graph object of the chosen backend, and the GEFF metadata.
    """  # noqa: E501
    backend_io = get_backend(backend)
    return backend_io.read(
        store,
        structure_validation,
        node_props,
        edge_props,
        data_validation,
        **backend_kwargs,
    )


@overload
def construct(
    metadata: GeffMetadata,
    node_ids: NDArray[Any],
    edge_ids: NDArray[Any],
    node_props: dict[str, PropDictNpArray],
    edge_props: dict[str, PropDictNpArray],
    *,
    backend: Literal["networkx"] = ...,
) -> NxGraph: ...
@overload
def construct(
    metadata: GeffMetadata,
    node_ids: NDArray[Any],
    edge_ids: NDArray[Any],
    node_props: dict[str, PropDictNpArray],
    edge_props: dict[str, PropDictNpArray],
    *,
    backend: Literal["rustworkx"],
) -> RxGraph: ...
@overload
def construct(
    metadata: GeffMetadata,
    node_ids: NDArray[Any],
    edge_ids: NDArray[Any],
    node_props: dict[str, PropDictNpArray],
    edge_props: dict[str, PropDictNpArray],
    *,
    backend: Literal["spatial-graph"],
    position_attr: str = "position",
) -> SgGraph: ...
def construct(
    metadata: GeffMetadata,
    node_ids: NDArray[Any],
    edge_ids: NDArray[Any],
    node_props: dict[str, PropDictNpArray],
    edge_props: dict[str, PropDictNpArray],
    *,
    backend: SupportedBackend = "networkx",
    **backend_kwargs: Any,
) -> SupportedGraphType:
    backend_io = get_backend(backend)
    return backend_io.construct(
        metadata, node_ids, edge_ids, node_props, edge_props, **backend_kwargs
    )


# rustworkx has an additional node_id_dict arg
@overload
def write(
    graph: RxGraph,
    store: StoreLike,
    metadata: GeffMetadata | None = ...,
    axis_names: list[str] | None = ...,
    axis_units: list[str | None] | None = ...,
    axis_types: list[AxisType | None] | None = ...,
    axis_scales: list[float | None] | None = ...,
    scaled_units: list[str | None] | None = ...,
    axis_offset: list[float | None] | None = ...,
    zarr_format: Literal[2, 3] = ...,
    structure_validation: bool = True,
    overwrite: bool = False,
    node_id_dict: dict[int, int] | None = ...,
) -> None:
    # TODO: what is best practice for overload docstrings, want to document node_id_dict
    """Write a rustworkx graph object to the geff file format.

    Args:
        graph (SupportedGraphType): An instance of a supported graph object.
        store (zarr.storage.StoreLike): The path/str to the output zarr, or the store
            itself. Opens in append mode, so will only overwrite geff-controlled groups.
        metadata (GeffMetadata, optional): The original metadata of the graph.
            Defaults to None. If provided, will override the graph properties.
        axis_names (list[str], optional): The names of the spatial dims
            represented in position property. Defaults to None. Will override
            both value in graph properties and metadata if provided.
        axis_units (list[str | None], optional): The units of the spatial dims
            represented in position property. Defaults to None. Will override value
            both value in graph properties and metadata if provided.
        axis_types (list[Literal[AxisType] | None], optional): The types of the spatial dims
            represented in position property. Usually one of "time", "space", or "channel".
            Defaults to None. Will override both value in graph properties and metadata
            if provided.
        axis_scales (list[float | None] | None): The scale to apply to the spatial dims.
            Defaults to None.
        scaled_units (list[str | None] | None): The units of the spatial dims after scaling.
            Defaults to None.
        axis_offset (list[float | None] | None): Amount to offset an axis after applying
            scaling factor. Defaults to None.
        zarr_format (Literal[2, 3], optional): The version of zarr to write.
            Defaults to 2.
        structure_validation (bool): If True, runs structural validation and does not write
            a geff that is invalid. Defaults to True.
        overwrite (bool): If True, deletes any existing geff and writes a new geff.
            Defaults to False.
        node_id_dict (dict[int, int], optional): A dictionary mapping rx node indices to
            arbitrary indices. This allows custom node identifiers to be used in the geff file
            instead of rustworkx's internal indices. If None, uses rx indices directly.
    """
    ...


@overload
def write(
    graph: SupportedGraphType,
    store: StoreLike,
    metadata: GeffMetadata | None = ...,
    axis_names: list[str] | None = ...,
    axis_units: list[str | None] | None = ...,
    axis_types: list[AxisType | None] | None = ...,
    axis_scales: list[float | None] | None = ...,
    scaled_units: list[str | None] | None = ...,
    axis_offset: list[float | None] | None = ...,
    zarr_format: Literal[2, 3] = ...,
    structure_validation: bool = True,
    overwrite: bool = False,
    *args: Any,
    **kwargs: Any,
) -> None: ...
def write(
    graph: SupportedGraphType,
    store: StoreLike,
    metadata: GeffMetadata | None = None,
    axis_names: list[str] | None = None,
    axis_units: list[str | None] | None = None,
    axis_types: list[AxisType | None] | None = None,
    axis_scales: list[float | None] | None = None,
    scaled_units: list[str | None] | None = None,
    axis_offset: list[float | None] | None = None,
    zarr_format: Literal[2, 3] = 2,
    structure_validation: bool = True,
    overwrite: bool = False,
    *args: Any,
    **kwargs: Any,
) -> None:
    """Write a supported graph object to the geff file format.

    Args:
        graph (networkx.Graph | networkx.DiGraph | rustworkx.PyGraph | rustworkx.PyDiGraph | spatial_graph.SpatialGraph | spatial_graph.SpatialDiGraph):
            An instance of a supported graph object.
        store (str | Path | zarr store): The path/str to the output zarr, or the store
            itself. Opens in append mode, so will only overwrite geff-controlled groups.
        metadata (GeffMetadata, optional): The original metadata of the graph.
            Defaults to None. If provided, will override the graph properties.
        axis_names (list[str], optional): The names of the spatial dims
            represented in position property. Defaults to None. Will override
            both value in graph properties and metadata if provided.
        axis_units (list[str | None], optional): The units of the spatial dims
            represented in position property. Defaults to None. Will override value
            both value in graph properties and metadata if provided.
        axis_types (list[AxisType | None] | None, optional): The types of the spatial dims
            represented in position property. Usually one of "time", "space", or "channel".
            Defaults to None. Will override both value in graph properties and metadata
            if provided.
        axis_scales (list[float | None] | None): The scale to apply to the spatial dims.
            Defaults to None.
        scaled_units (list[str | None] | None): The units of the spatial dims after scaling.
            Defaults to None.
        axis_offset (list[float | None] | None): Amount to offset an axis after applying
            scaling factor. Defaults to None.
        zarr_format (Literal[2, 3], optional): The version of zarr to write.
            Defaults to 2.
        structure_validation (bool): If True, runs structural validation and does not write
            a geff that is invalid. Defaults to True.
        overwrite (bool): If True, deletes any existing geff and writes a new geff.
            Defaults to False.
        *args (Any): Additional args that may be accepted by the backend when writing from a
            specific type of graph.
        **kwargs (Any): Additional kwargs that may be accepted by the backend when writing from a
            specific type of graph.
    """  # noqa: E501
    store = remove_tilde(store)

    # Check for existing geff
    if check_for_geff(store):
        if overwrite:
            delete_geff(store, zarr_format=zarr_format)
        else:
            raise FileExistsError(
                "Found an existing geff present in `store`. "
                "Please use `overwrite=True` or provide an alternative "
                "`store` to write to."
            )

    backend_io = get_backend_from_graph_type(graph)
    backend_io.write(
        graph,
        store,
        metadata,
        axis_names,
        axis_units,
        axis_types,
        axis_scales,
        scaled_units,
        axis_offset,
        zarr_format,
        structure_validation,
        *args,
        **kwargs,
    )
